# -*- coding: utf-8 -*-

"""Convert KEGG Genes to OBO.

Run with ``python -m pyobo.sources.kegg.genes``
"""

import logging
from typing import Iterable, Optional

import click
from more_click import verbose_option
from tqdm.auto import tqdm

from .api import (
    KEGG_GENES_PREFIX,
    SKIP,
    KEGGGenome,
    ensure_conv_genome_ncbigene,
    ensure_conv_genome_uniprot,
    ensure_list_genome,
)
from .genome import iter_kegg_genomes
from ...struct import Obo, Reference, Term, from_species, has_gene_product
from ...utils.io import open_map_tsv

__all__ = [
    "KEGGGeneGetter",
]

logger = logging.getLogger(__name__)


class KEGGGeneGetter(Obo):
    """An ontology representation of KEGG Genes."""

    ontology = KEGG_GENES_PREFIX
    bioversions_key = "kegg"
    typedefs = [from_species, has_gene_product]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise)


def get_obo() -> Obo:
    """Get KEGG Genes as OBO."""
    return KEGGGeneGetter()


def iter_terms(version: str) -> Iterable[Term]:
    """Iterate over terms for KEGG Genome."""
    for kegg_genome in iter_kegg_genomes(version=version, desc="KEGG Genes"):
        if kegg_genome.identifier in SKIP:
            continue
        try:
            list_genome_path = ensure_list_genome(kegg_genome.identifier, version=version)
            conv_uniprot_path = ensure_conv_genome_uniprot(kegg_genome.identifier, version=version)
            conv_ncbigene_path = ensure_conv_genome_ncbigene(
                kegg_genome.identifier, version=version
            )
        except (OSError, ValueError) as e:
            tqdm.write(f"[{KEGG_GENES_PREFIX}] {kegg_genome.identifier} failed: {e}")
            continue

        yield from _make_terms(
            kegg_genome,
            list_genome_path,
            conv_uniprot_path,
            conv_ncbigene_path,
        )


def _make_terms(
    kegg_genome: KEGGGenome,
    list_genome_path: str,
    conv_uniprot_path: Optional[str] = None,
    conv_ncbigene_path: Optional[str] = None,
) -> Iterable[Term]:
    uniprot_conv = _load_conv(conv_uniprot_path, "up:") if conv_uniprot_path else {}
    ncbigene_conv = _load_conv(conv_ncbigene_path, "ncbi-geneid:") if conv_ncbigene_path else {}

    with open(list_genome_path) as file:
        for line in file:
            try:
                identifier, extras = line.strip().split("\t")
            except ValueError:
                logger.warning(
                    "[%s] could not parse line in %s: %s", KEGG_GENES_PREFIX, list_genome_path, line
                )
                continue
            if ";" in line:
                *_extras, name = [part.strip() for part in extras.split(";")]
            else:
                name = extras

            term = Term.from_triple(
                prefix=KEGG_GENES_PREFIX,
                identifier=identifier,
                name=name,
            )

            uniprot_xref = uniprot_conv.get(identifier)
            if uniprot_xref is not None:
                term.append_relationship(has_gene_product, Reference("uniprot", uniprot_xref))

            ncbigene_xref = ncbigene_conv.get(identifier)
            if ncbigene_xref is not None:
                term.append_xref(Reference(prefix="ncbigene", identifier=ncbigene_xref))

            kegg_genome.annotate_term(term)
            yield term


def _load_conv(path, value_prefix):
    m = open_map_tsv(path)
    m = {k: v[len(value_prefix) :] for k, v in m.items()}
    return m


@click.command()
@verbose_option
def _main():
    get_obo().write_default()


if __name__ == "__main__":
    _main()



# -*- coding: utf-8 -*-

"""Utilities for handling prefixes."""

import logging
from functools import wraps
from typing import Optional, Tuple, Union

import bioregistry

from .registries import (
    curie_has_blacklisted_prefix,
    curie_has_blacklisted_suffix,
    curie_is_blacklisted,
    remap_full,
    remap_prefix,
)

__all__ = [
    "normalize_curie",
    "wrap_norm_prefix",
    "standardize_ec",
]

logger = logging.getLogger(__name__)


class MissingPrefix(ValueError):
    """Raised on a missing prefix."""

    def __init__(self, prefix, curie, xref=None, ontology=None):
        self.prefix = prefix
        self.curie = curie
        self.xref = xref
        self.ontology = ontology
        self.reference = None

    def __str__(self) -> str:
        s = ""
        if self.ontology:
            s += f"[{self.ontology}] "
        s += f"unhandled prefix {self.prefix} found in curie {self.curie}"
        if self.xref:
            s += f"/xref {self.xref}"
        if self.reference is not None:
            s += f" from {self.reference.curie}"
        return s


def _normalize_prefix(prefix: str, *, curie=None, xref=None, strict: bool = True) -> Optional[str]:
    """Normalize a namespace and return, if possible."""
    norm_prefix = bioregistry.normalize_prefix(prefix)
    if norm_prefix is not None:
        return norm_prefix
    elif strict:
        raise MissingPrefix(prefix=prefix, curie=curie, xref=xref)
    else:
        return None


BAD_CURIES = set()


def normalize_curie(
    curie: str, *, strict: bool = True
) -> Union[Tuple[str, str], Tuple[None, None]]:
    """Parse a string that looks like a CURIE.

    :param curie: A compact uniform resource identifier (CURIE)
    :param strict: Should an exception be thrown if the CURIE can not be parsed w.r.t. the Bioregistry?
    :return: A parse tuple or a tuple of None, None if not able to parse and not strict

    - Normalizes the namespace
    - Checks against a blacklist for the entire curie, for the namespace, and for suffixes.
    """
    if curie_is_blacklisted(curie):
        return None, None
    if curie_has_blacklisted_prefix(curie):
        return None, None
    if curie_has_blacklisted_suffix(curie):
        return None, None

    # Remap the curie with the full list
    curie = remap_full(curie)

    # Remap node's prefix (if necessary)
    curie = remap_prefix(curie)

    try:
        head_ns, identifier = curie.split(":", 1)
    except ValueError:  # skip nodes that don't look like normal CURIEs
        if curie not in BAD_CURIES:
            BAD_CURIES.add(curie)
            logger.debug(f"could not split CURIE on colon: {curie}")
        return None, None

    # remove redundant prefix
    if identifier.casefold().startswith(f"{head_ns.casefold()}:"):
        identifier = identifier[len(head_ns) + 1 :]

    norm_node_prefix = _normalize_prefix(head_ns, curie=curie, strict=strict)
    if not norm_node_prefix:
        return None, None
    return norm_node_prefix, identifier


def wrap_norm_prefix(f):
    """Decorate a function that take in a prefix to auto-normalize, or return None if it can't be normalized."""

    @wraps(f)
    def _wrapped(prefix, *args, **kwargs):
        norm_prefix = bioregistry.normalize_prefix(prefix)
        if norm_prefix is None:
            raise ValueError(f"Invalid prefix: {prefix}")
        return f(norm_prefix, *args, **kwargs)

    return _wrapped


def standardize_ec(ec: str) -> str:
    """Standardize an EC code identifier by removing all trailing dashes and dots."""
    ec = ec.strip()
    for _ in range(4):
        ec = ec.rstrip("-").rstrip(".")
    return ec



# -*- coding: utf-8 -*-

"""High-level API for species."""

import logging
from functools import lru_cache
from typing import Mapping, Optional

from .alts import get_primary_identifier
from .utils import get_version
from ..getters import NoBuild, get_ontology
from ..identifier_utils import wrap_norm_prefix
from ..utils.cache import cached_mapping
from ..utils.path import prefix_cache_join

__all__ = [
    "get_id_species_mapping",
    "get_species",
]

logger = logging.getLogger(__name__)


@wrap_norm_prefix
def get_species(prefix: str, identifier: str) -> Optional[str]:
    """Get the species."""
    if prefix == "uniprot":
        raise NotImplementedError

    try:
        id_species = get_id_species_mapping(prefix)
    except NoBuild:
        logger.warning("unable to look up species for prefix %s", prefix)
        return None

    if not id_species:
        logger.warning("no results produced for prefix %s", prefix)
        return None

    primary_id = get_primary_identifier(prefix, identifier)
    return id_species.get(primary_id)


@lru_cache()
@wrap_norm_prefix
def get_id_species_mapping(
    prefix: str,
    force: bool = False,
    strict: bool = True,
    version: Optional[str] = None,
) -> Mapping[str, str]:
    """Get an identifier to species mapping."""
    if prefix == "ncbigene":
        from ..sources.ncbigene import get_ncbigene_id_to_species_mapping

        logger.info("[%s] loading species mappings", prefix)
        rv = get_ncbigene_id_to_species_mapping()
        logger.info("[%s] done loading species mappings", prefix)
        return rv

    if version is None:
        version = get_version(prefix)
    path = prefix_cache_join(prefix, name="species.tsv", version=version)

    @cached_mapping(path=path, header=[f"{prefix}_id", "species"], force=force)
    def _get_id_species_mapping() -> Mapping[str, str]:
        logger.info("[%s] no cached species found. getting from OBO loader", prefix)
        ontology = get_ontology(prefix, force=force, strict=strict, version=version)
        logger.info("[%s] loading species mappings", prefix)
        return ontology.get_id_species_mapping()

    return _get_id_species_mapping()



# -*- coding: utf-8 -*-

"""Constants for miRBase."""

import pandas as pd

from pyobo.utils.path import ensure_df

PREFIX = "mirbase"

# PREMATURE_TO_MATURE = f'https://mirbase.org/ftp/{VERSION}/database_files/mirna_pre_mature.txt.gz'


def get_premature_family_df(version: str, force: bool = False) -> pd.DataFrame:
    """Get premature family dataframe."""
    url = f"https://mirbase.org/ftp/{version}/database_files/mirna_prefam.txt.gz"
    return ensure_df(
        PREFIX,
        url=url,
        version=version,
        names=["prefamily_key", "family_id", "family_name"],
        usecols=[0, 1, 2],
        dtype=str,
        force=force,
    )


def get_premature_to_prefamily_df(version: str, force: bool = False) -> pd.DataFrame:
    """Get premature miRNA to premature family dataframe."""
    url = f"https://mirbase.org/ftp/{version}/database_files/mirna_2_prefam.txt.gz"
    return ensure_df(
        PREFIX,
        url=url,
        version=version,
        names=["premature_key", "prefamily_key"],
        dtype=str,
        force=force,
    )


def get_premature_df(version: str, force: bool = False) -> pd.DataFrame:
    """Get premature miRNA dataframe."""
    url = f"https://mirbase.org/ftp/{version}/database_files/mirna.txt.gz"
    return ensure_df(
        PREFIX,
        url=url,
        version=version,
        names=["premature_key", "mirbase_id", "mirna_name"],
        usecols=[0, 1, 2],
        dtype=str,
        force=force,
    )


def get_mature_df(version: str, force: bool = False) -> pd.DataFrame:
    """Get mature miRNA dataframe."""
    url = f"https://mirbase.org/ftp/{version}/database_files/mirna_mature.txt.gz"
    return ensure_df(
        PREFIX,
        url=url,
        version=version,
        names=[
            "mature_key",
            "name",
            "previous",
            "mirbase.mature_id",
        ],
        usecols=[0, 1, 2, 3],
        dtype=str,
        force=force,
    )



# -*- coding: utf-8 -*-

"""Converter for ZFIN."""

import logging
from collections import defaultdict
from typing import Iterable, Optional

from tqdm.auto import tqdm

from pyobo.struct import (
    Obo,
    Reference,
    Term,
    from_species,
    has_gene_product,
    orthologous,
)
from pyobo.utils.io import multidict, multisetdict
from pyobo.utils.path import ensure_df

__all__ = [
    "ZFINGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "zfin"
NAME = "Zebrafish Information Network"
URL = "https://zfin.org/downloads/genetic_markers.txt"
ALTS_URL = "https://zfin.org/downloads/zdb_history.txt"
HUMAN_ORTHOLOGS = "https://zfin.org/downloads/human_orthos.txt"
MOUSE_ORTHOLOGS = "https://zfin.org/downloads/mouse_orthos.txt"
FLY_ORTHOLOGS = "https://zfin.org/downloads/fly_orthos.txt"
ENTREZ_MAPPINGS = "https://zfin.org/downloads/gene.txt"
UNIPROT_MAPPINGS = "https://zfin.org/downloads/uniprot.txt"


class ZFINGetter(Obo):
    """An ontology representation of ZFIN's zebrafish database."""

    bioversions_key = ontology = PREFIX
    typedefs = [from_species, has_gene_product, orthologous]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in ZFIN."""
        return get_terms(force=force, version=self._version_or_raise)


def get_obo(force: bool = False) -> Obo:
    """Get ZFIN OBO."""
    return ZFINGetter(force=force)


MARKERS_COLUMNS = [
    "zfin_id",
    "name",
    "description",
    "supertype",
    "sequence_ontology_id",
]


def get_terms(force: bool = False, version: Optional[str] = None) -> Iterable[Term]:
    """Get terms."""
    alt_ids_df = ensure_df(
        PREFIX,
        url=ALTS_URL,
        name="alts.tsv",
        force=force,
        header=None,
        names=["alt", "zfin_id"],
        version=version,
    )
    primary_to_alt_ids = defaultdict(set)
    for alt_id, zfin_id in alt_ids_df.values:
        primary_to_alt_ids[zfin_id].add(alt_id)

    human_orthologs = multisetdict(
        ensure_df(
            PREFIX, url=HUMAN_ORTHOLOGS, force=force, header=None, usecols=[0, 7], version=version
        ).values
    )
    mouse_orthologs = multisetdict(
        ensure_df(
            PREFIX, url=MOUSE_ORTHOLOGS, force=force, header=None, usecols=[0, 5], version=version
        ).values
    )
    fly_orthologs = multisetdict(
        ensure_df(
            PREFIX, url=FLY_ORTHOLOGS, force=force, header=None, usecols=[0, 5], version=version
        ).values
    )
    entrez_mappings = dict(
        ensure_df(
            PREFIX, url=ENTREZ_MAPPINGS, force=force, header=None, usecols=[0, 3], version=version
        ).values
    )
    uniprot_mappings = multidict(
        ensure_df(
            PREFIX, url=UNIPROT_MAPPINGS, force=force, header=None, usecols=[0, 3], version=version
        ).values
    )

    df = ensure_df(
        PREFIX,
        url=URL,
        name="markers.tsv",
        force=force,
        header=None,
        names=MARKERS_COLUMNS,
        version=version,
    )
    df["sequence_ontology_id"] = df["sequence_ontology_id"].map(lambda x: x[len("SO:") :])
    so = {
        sequence_ontology_id: Reference.auto(prefix="SO", identifier=sequence_ontology_id)
        for sequence_ontology_id in df["sequence_ontology_id"].unique()
    }
    for _, reference in sorted(so.items()):
        yield Term(reference=reference)
    for identifier, name, definition, _entity_type, sequence_ontology_id in tqdm(
        df.values, unit_scale=True, unit="gene", desc="zfin"
    ):
        term = Term.from_triple(
            prefix=PREFIX,
            identifier=identifier,
            name=name,
            definition=definition if definition != name else None,
        )
        term.set_species(identifier="7955", name="Danio rerio")
        term.append_parent(so[sequence_ontology_id])
        # Entity type is redundant of identifier
        # term.append_property("type", entity_type)
        for alt_id in primary_to_alt_ids[identifier]:
            term.append_alt(alt_id)
        entrez_id = entrez_mappings.get(identifier)
        if entrez_id:
            term.append_xref(Reference(prefix="ncbigene", identifier=entrez_id))
        for uniprot_id in uniprot_mappings.get(identifier, []):
            term.append_relationship(has_gene_product, Reference.auto("uniprot", uniprot_id))
        for hgnc_id in human_orthologs.get(identifier, []):
            term.append_relationship(orthologous, Reference.auto("hgnc", hgnc_id))
        for mgi_curie in mouse_orthologs.get(identifier, []):
            mouse_ortholog = Reference.from_curie(mgi_curie, auto=True)
            if mouse_ortholog:
                term.append_relationship(orthologous, mouse_ortholog)
        for flybase_id in fly_orthologs.get(identifier, []):
            term.append_relationship(orthologous, Reference.auto("flybase", flybase_id))

        yield term


if __name__ == "__main__":
    ZFINGetter.cli()



"""Converter for UniProt PTMs.

Line code  Content                         Occurrence in an entry
---------  ---------------------------     ------------------------------
ID         Identifier (FT description)     Once; starts a PTM entry
AC         Accession (PTM-xxxx)            Once
FT         Feature key                     Once
TG         Target                          Once; two targets separated by
                                           a dash in case of intrachain
                                           crosslinks
PA         Position of the modification    Optional; once
           on the amino acid
PP         Position of the modification    Optional; once
           in the polypeptide
CF         Correction formula              Optional; once
MM         Monoisotopic mass difference    Optional; once
MA         Average mass difference         Optional; once
LC         Cellular location               Optional; once; alternatives
                                           can be proposed
TR         Taxonomic range                 Optional; once or more
KW         Keyword                         Optional; once or more
DR         Cross-reference to external     Optional; once or more
           databases
//         Terminator                      Once; ends an entr

"""

import itertools as itt
from collections import defaultdict
from typing import DefaultDict, Iterable, List, Mapping, Optional, Tuple

from tqdm.auto import tqdm

from pyobo import Obo, Reference, Term
from pyobo.utils.path import ensure_path

__all__ = [
    "UniProtPtmGetter",
]

PREFIX = "uniprot.ptm"
URL = "https://www.uniprot.org/docs/ptmlist.txt"


class UniProtPtmGetter(Obo):
    """An ontology representation of the UniProt PTM database."""

    ontology = PREFIX
    dynamic_version = True

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        yield from iter_terms(force=force)


def get_obo(force: bool = False) -> Obo:
    """Get UniProt PTMs as OBO."""
    return UniProtPtmGetter(force=force)


def iter_terms(force: bool = False) -> Iterable[Term]:
    """Iterate over UniProt PTM Terms."""
    path = ensure_path(PREFIX, url=URL, force=force)
    with open(path) as file:
        lines = list(file)
    it: Iterable[Tuple[str, str]] = ((line[:2], line[2:].strip()) for line in lines[47:-5])
    for i, (_, term_lines) in enumerate(itt.groupby(it, key=lambda p: p[0] == "//")):
        term = _parse(i, term_lines)
        if term:
            yield term


def _parse(i, lines: Iterable[Tuple[str, str]]) -> Optional[Term]:
    dd_: DefaultDict[str, List[str]] = defaultdict(list)
    for key, value in lines:
        dd_[key].append(value)
    dd: Mapping[str, List[str]] = dict(dd_)

    if "//" in dd:
        return None

    accessions = dd["AC"]
    labels = dd.get("ID")
    reference = Reference(
        prefix="uniprot.ptm",
        identifier=accessions[0],
        name=labels[0] if labels else None,
    )
    term = Term(reference=reference)
    for line in dd.get("DR", []):
        line = line.rstrip(".")
        for x, y in [
            ("MOD; ", "PSI-MOD; MOD:"),
            ("CHEBI; ", "ChEBI; CHEBI:"),
        ]:
            if line.startswith(y):
                line = x + line[len(y) :]

        ref = Reference.from_curie(line.replace("; ", ":"))
        if ref:
            term.append_xref(ref)
        else:
            tqdm.write(f"Failure on xref {line}")
    return term


if __name__ == "__main__":
    UniProtPtmGetter.cli()



# -*- coding: utf-8 -*-

"""Utilities for converters."""

import logging
from typing import Mapping, Set, Tuple

from ..utils.io import multisetdict

__all__ = [
    "get_go_mapping",
    "process_go_mapping_line",
]

logger = logging.getLogger(__name__)


def get_go_mapping(path: str, prefix: str) -> Mapping[str, Set[Tuple[str, str]]]:
    """Get a GO mapping file."""
    with open(path) as file:
        return multisetdict(
            process_go_mapping_line(line.strip(), prefix=prefix) for line in file if line[0] != "!"
        )


def process_go_mapping_line(line: str, prefix: str) -> Tuple[str, Tuple[str, str]]:
    """Process a GO mapping line."""
    line1 = line[len(f"{prefix}:") :]
    line2, go_id = line1.rsplit(";", 1)
    go_id = go_id.strip()[len("GO:") :]
    try:
        line3, go_name = line2.rsplit(">", 1)
    except ValueError:
        logger.warning(line)
        logger.warning("go:%s", go_id)
        raise

    go_name = go_name.strip()[len("GO:") :]
    interpro_id, _interpro_name = line3.split(" ", 1)
    interpro_id = interpro_id.strip()
    return interpro_id, (go_id, go_name)



# -*- coding: utf-8 -*-

"""High-level API for metadata."""

import logging
from functools import lru_cache
from typing import Mapping

from .utils import get_version
from ..getters import get_ontology
from ..identifier_utils import wrap_norm_prefix
from ..utils.cache import cached_json
from ..utils.path import prefix_cache_join

__all__ = [
    "get_metadata",
]

logger = logging.getLogger(__name__)


@lru_cache()
@wrap_norm_prefix
def get_metadata(prefix: str, force: bool = False) -> Mapping[str, str]:
    """Get metadata for the ontology."""
    version = get_version(prefix)
    path = prefix_cache_join(prefix, name="metadata.json", version=version)

    @cached_json(path=path, force=force)
    def _get_json() -> Mapping[str, str]:
        if force:
            logger.debug("[%s] forcing reload for metadata", prefix)
        else:
            logger.debug("[%s] no cached metadata found. getting from OBO loader", prefix)
        ontology = get_ontology(prefix, force=force, version=version)
        return ontology.get_metadata()

    return _get_json()



# -*- coding: utf-8 -*-

"""Converter for PomBase."""

import logging
from collections import defaultdict
from typing import Iterable

import pandas as pd
from tqdm.auto import tqdm

import pyobo
from pyobo import Reference
from pyobo.struct import Obo, Term, from_species, has_gene_product, orthologous
from pyobo.utils.path import ensure_df

__all__ = [
    "PomBaseGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "pombase"
URL = "https://www.pombase.org/data/names_and_identifiers/gene_IDs_names_products.tsv"
ORTHOLOGS_URL = "https://www.pombase.org/data/orthologs/human-orthologs.txt.gz"


class PomBaseGetter(Obo):
    """An ontology representation of PomBase's fission yeast gene nomenclature."""

    ontology = bioversions_key = PREFIX
    typedefs = [from_species, has_gene_product]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return get_terms(force=force, version=self._version_or_raise)


def get_obo(force: bool = False) -> Obo:
    """Get OBO."""
    return PomBaseGetter(force=force)


#: A mapping from PomBase gene type to sequence ontology terms
POMBASE_TO_SO = {
    # None: "0000704",  # gene,
    "gene_type": "0000704",  # unannotated
    "protein coding gene": "0001217",
    "pseudogene": "0000336",
    "tRNA gene": "0001272",
    "ncRNA gene": "0001263",
    "snRNA gene": "0001268",
    "snoRNA gene": "0001267",
    "rRNA gene": "0001637",
    "lncRNA gene": "0002127",
    "sncRNA gene": "0002342",
}


def get_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Get terms."""
    orthologs_df = ensure_df(PREFIX, url=ORTHOLOGS_URL, force=force, header=None, version=version)
    identifier_to_hgnc_ids = defaultdict(set)
    hgnc_symbol_to_id = pyobo.get_name_id_mapping("hgnc")
    for identifier, hgnc_symbols in orthologs_df.values:
        if hgnc_symbols == "NONE":
            continue
        for hgnc_symbol in hgnc_symbols.split("|"):
            hgnc_id = hgnc_symbol_to_id.get(hgnc_symbol)
            if hgnc_id is not None:
                identifier_to_hgnc_ids[identifier].add(hgnc_id)

    df = ensure_df(PREFIX, url=URL, force=force, header=None, version=version)
    so = {
        gtype: Reference.auto("SO", POMBASE_TO_SO[gtype])
        for gtype in sorted(df[df.columns[6]].unique())
    }
    for _, reference in sorted(so.items()):
        yield Term(reference=reference)
    for identifier, _, symbol, chromosome, name, uniprot_id, gtype, synonyms in tqdm(
        df.values, unit_scale=True
    ):
        if pd.isna(identifier):
            continue
        term = Term.from_triple(
            prefix=PREFIX,
            identifier=identifier,
            name=symbol if pd.notna(symbol) else None,
            definition=name if pd.notna(name) else None,
        )
        term.append_property("chromosome", chromosome[len("chromosome_") :])
        term.append_parent(so[gtype])
        term.set_species(identifier="4896", name="Schizosaccharomyces pombe")
        for hgnc_id in identifier_to_hgnc_ids.get(identifier, []):
            term.append_relationship(orthologous, Reference(prefix="hgnc", identifier=hgnc_id))
        if uniprot_id and pd.notna(uniprot_id):
            term.append_relationship(
                has_gene_product, Reference(prefix="uniprot", identifier=uniprot_id)
            )
        if synonyms and pd.notna(synonyms):
            for synonym in synonyms.split(","):
                term.append_synonym(synonym.strip())
        yield term


if __name__ == "__main__":
    PomBaseGetter.cli()



# -*- coding: utf-8 -*-

"""Converters for UniProt resources."""

from .uniprot import PREFIX, UniProtGetter
from .uniprot_ptm import UniProtPtmGetter

__all__ = [
    "UniProtGetter",
    "UniProtPtmGetter",
    "PREFIX",
]



# -*- coding: utf-8 -*-

"""Converter for GWAS Central."""

import logging
import tarfile
from typing import Iterable, Optional
from xml.etree import ElementTree

from pyobo.struct import Obo, Reference, Term, has_part
from pyobo.utils.path import ensure_path

__all__ = [
    "GWASCentralStudyGetter",
]

logger = logging.getLogger(__name__)

VERSION = "jan2021"
PREFIX = "gwascentral.study"


class GWASCentralStudyGetter(Obo):
    """An ontology representation of GWAS Central's study nomenclature."""

    ontology = PREFIX
    static_version = VERSION
    typedefs = [has_part]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iterate_terms(force=force, version=self._version_or_raise)


def get_obo(force: bool = False):
    """Get GWAS Central Studies as OBO."""
    return GWASCentralStudyGetter(force=force)


def _find_text(element, name: str) -> Optional[str]:
    x = element.find(name)
    if x is not None:
        return x.text
    return None


def _get_term_from_tree(tree: ElementTree.ElementTree) -> Term:
    name = _find_text(tree, "name")
    description = _find_text(tree, "description")
    if description:
        description = description.strip().replace("\n", " ")
    identifier = _find_text(tree, "identifier")
    if identifier is None:
        raise ValueError
    term = Term(
        reference=Reference(prefix=PREFIX, identifier=identifier, name=name),
        definition=description,
    )
    for experiment in tree.findall("experiments"):
        experiment_name = _find_text(experiment, "name")
        experiment_identifier = _find_text(experiment, "identifier")
        if experiment_identifier is None:
            continue
        term.append_relationship(
            has_part,
            Reference(
                prefix="gwascentral.experiment",
                identifier=experiment_identifier,
                name=experiment_name,
            ),
        )
    return term


def iterate_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Iterate over GWAS Central Study terms."""
    url = f"http://www.gwascentral.org/docs/GC_{version}.tar.gz"
    path = ensure_path(PREFIX, url=url, version=version, force=force)
    with tarfile.open(path) as tar_file:
        for tar_info in tar_file:
            if not tar_info.path.endswith(".xml"):
                continue
            with tar_file.extractfile(tar_info) as file:  # type:ignore
                try:
                    tree = ElementTree.parse(file)
                except ElementTree.ParseError:
                    logger.warning("malformed XML in %s", tar_info.path)
                    continue
            yield _get_term_from_tree(tree)


if __name__ == "__main__":
    GWASCentralStudyGetter.cli()



# -*- coding: utf-8 -*-

"""CLI for PyOBO's Mapping Service.

Run with ``python -m pyobo.apps.mapper``.
"""

import click
from more_click import (
    host_option,
    port_option,
    run_app,
    verbose_option,
    with_gunicorn_option,
)

__all__ = [
    "main",
]


@click.command(name="mapper")
@click.option("-x", "--mappings-file")
@port_option
@host_option
@with_gunicorn_option
@verbose_option
def main(mappings_file, host: str, port: str, with_gunicorn: bool):
    """Run the mappings app."""
    from .mapper import get_app

    app = get_app(mappings_file)
    run_app(app=app, host=host, port=port, with_gunicorn=with_gunicorn)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Converter for PathBank."""

import logging
from collections import defaultdict
from typing import Iterable, Mapping, Set

import pandas as pd
from tqdm.auto import tqdm

from ..struct import Obo, Reference, Term
from ..struct.typedef import has_participant
from ..utils.path import ensure_df

__all__ = [
    "PathBankGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "pathbank"

PATHWAY_URL = "https://pathbank.org/downloads/pathbank_all_pathways.csv.zip"
PATHWAY_COLUMNS = [
    "SMPDB ID",
    "PW ID",
    "Name",
    "Subject",
    "Description",
]

PROTEINS_URL = "https://pathbank.org/downloads/pathbank_all_proteins.csv.zip"
PROTEINS_COLUMNS = [
    "PathBank ID",
    "Pathway Name",
    "Pathway Subject",
    "Species",
    "Uniprot ID",
    "Protein Name",
    "HMDBP ID",
    "DrugBank ID",
    "GenBank ID",
    "Gene Name",
    "Locus",
]

METABOLITE_URL = "https://pathbank.org/downloads/pathbank_all_metabolites.csv.zip"
METABOLITE_COLUMNS = [
    "PathBank ID",
    "Pathway Name",
    "Pathway Subject",
    "Species",
    "Metabolite ID",
    "Metabolite Name",
    "HMDB ID",
    "KEGG ID",
    "ChEBI ID",
    "DrugBank ID",
    "CAS",
    "Formula",
    "IUPAC",
    "SMILES",
    "InChI",
    "InChI Key",
]


class PathBankGetter(Obo):
    """An ontology representation of PathBank's pathway nomenclature."""

    ontology = bioversions_key = PREFIX
    typedefs = [has_participant]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(force=force, version=self._version_or_raise)


def get_obo(force: bool = False) -> Obo:
    """Get PathBank as OBO."""
    return PathBankGetter(force=force)


def get_proteins_df(version: str, force: bool = False) -> pd.DataFrame:
    """Get the proteins dataframe."""
    proteins_df = ensure_df(
        PREFIX,
        url=PROTEINS_URL,
        sep=",",
        usecols=["PathBank ID", "Uniprot ID"],
        version=version,
        force=force,
    )
    proteins_df = proteins_df[proteins_df["Uniprot ID"].notna()]
    proteins_df = proteins_df[proteins_df["Uniprot ID"] != "Unknown"]
    proteins_df["Uniprot ID"] = proteins_df["Uniprot ID"].map(str.strip)
    return proteins_df


def get_protein_mapping(version: str, force: bool = False) -> Mapping[str, Set[Reference]]:
    """Make the protein mapping."""
    proteins_df = get_proteins_df(version=version, force=force)
    smpdb_id_to_proteins = defaultdict(set)
    for pathway_id, protein_id in tqdm(
        proteins_df.values, desc=f"[{PREFIX}] mapping proteins", unit_scale=True
    ):
        # TODO get protein names
        smpdb_id_to_proteins[pathway_id].add(Reference(prefix="uniprot", identifier=protein_id))
    return smpdb_id_to_proteins


def get_metabolite_df(version: str, force: bool = False) -> pd.DataFrame:
    """Get the metabolites dataframe."""
    return ensure_df(
        PREFIX,
        url=METABOLITE_URL,
        sep=",",
        usecols=["PathBank ID", "Metabolite ID", "Metabolite Name"],
        force=force,
        version=version,
    )


def get_metabolite_mapping(version: str, force: bool = False) -> Mapping[str, Set[Reference]]:
    """Make the metabolite mapping."""
    metabolites_df = get_metabolite_df(version=version, force=force)
    smpdb_id_to_metabolites = defaultdict(set)
    it = tqdm(metabolites_df.values, desc=f"[{PREFIX}] mapping metabolites", unit_scale=True)
    for pathway_id, metabolite_id, metabolite_name in it:
        smpdb_id_to_metabolites[pathway_id].add(
            Reference(
                prefix=PREFIX,
                identifier=metabolite_id,
                name=metabolite_name,
            )
        )
    return smpdb_id_to_metabolites


def iter_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Get PathBank's terms."""
    smpdb_id_to_proteins = get_protein_mapping(version=version, force=force)
    smpdb_id_to_metabolites = get_metabolite_mapping(version=version, force=force)

    pathways_df = ensure_df(PREFIX, url=PATHWAY_URL, sep=",", version=version, force=force)
    it = tqdm(pathways_df.values, total=len(pathways_df.index), desc=f"mapping {PREFIX}")
    for smpdb_id, pathbank_id, name, subject, _description in it:
        reference = Reference(prefix=PREFIX, identifier=pathbank_id, name=name)
        term = Term(
            reference=reference,
            # definition=description.replace('\n', ' '),
            xrefs=[Reference(prefix="smpdb", identifier=smpdb_id)],
        )
        term.append_parent(
            Reference(
                prefix=PREFIX,
                identifier=subject.lower().replace(" ", "_"),
                name=subject,
            )
        )
        term.extend_relationship(has_participant, smpdb_id_to_proteins[smpdb_id])
        term.extend_relationship(has_participant, smpdb_id_to_metabolites[smpdb_id])
        yield term


if __name__ == "__main__":
    PathBankGetter.cli()



# -*- coding: utf-8 -*-

"""GMT tests."""

import unittest

from pyobo.sources.gmt_utils import parse_gmt_file, parse_wikipathways_gmt
from tests.constants import TEST_GMT_PATH, TEST_WP_GMT_PATH


class TestGMT(unittest.TestCase):
    """Test parsing GMT files."""

    def test_parse_standard(self):
        """Test parsing a standard GMT file."""
        x = list(parse_gmt_file(TEST_GMT_PATH))
        self.assertEqual(3, len(x))

        self.assertEqual("HALLMARK_TNFA_SIGNALING_VIA_NFKB", x[0][0])
        self.assertEqual(
            "http://www.gsea-msigdb.org/gsea/msigdb/cards/HALLMARK_TNFA_SIGNALING_VIA_NFKB", x[0][1]
        )
        self.assertEqual({"3726", "2920"}, x[0][2])

        self.assertEqual("HALLMARK_HYPOXIA", x[1][0])
        self.assertEqual("http://www.gsea-msigdb.org/gsea/msigdb/cards/HALLMARK_HYPOXIA", x[1][1])
        self.assertEqual({"5230", "5163", "2632"}, x[1][2])

        self.assertEqual("HALLMARK_CHOLESTEROL_HOMEOSTASIS", x[2][0])
        self.assertEqual(
            "http://www.gsea-msigdb.org/gsea/msigdb/cards/HALLMARK_CHOLESTEROL_HOMEOSTASIS", x[2][1]
        )
        self.assertEqual({"2224", "1595"}, x[2][2])

    def test_parse_wikipathways(self):
        """Test parsing a WikiPathways GMT file."""
        x = list(parse_wikipathways_gmt(TEST_WP_GMT_PATH))
        self.assertEqual(3, len(x))

        self.assertEqual("WP4400", x[0][0])
        self.assertEqual("20200310", x[0][1])
        self.assertEqual("108112", x[0][2])
        self.assertEqual("FABP4 in ovarian cancer", x[0][3])
        self.assertEqual("Homo sapiens", x[0][4])
        self.assertEqual({"574413", "2167"}, x[0][5])

        self.assertEqual("WP23", x[1][0])
        self.assertEqual("20200310", x[1][1])
        self.assertEqual("108321", x[1][2])
        self.assertEqual("B Cell Receptor Signaling Pathway", x[1][3])
        self.assertEqual("Homo sapiens", x[1][4])
        self.assertEqual({"4690", "5781", "11184", "6195"}, x[1][5])

        self.assertEqual("WP2333", x[2][0])
        self.assertEqual("20200310", x[2][1])
        self.assertEqual("72015", x[2][2])
        self.assertEqual("Trans-sulfuration pathway", x[2][3])
        self.assertEqual("Homo sapiens", x[2][4])
        self.assertEqual({"1786", "2730", "27430"}, x[2][5])



# -*- coding: utf-8 -*-

"""Parsers for MSig."""

import logging
from typing import Iterable, Optional
from xml.etree import ElementTree

from tqdm.auto import tqdm

from ..struct import Obo, Reference, Term, has_participant
from ..utils.path import ensure_path

logger = logging.getLogger(__name__)

__all__ = [
    "MSigDBGetter",
]

PREFIX = "msigdb"
BASE_URL = "https://data.broadinstitute.org/gsea-msigdb/msigdb/release"


class MSigDBGetter(Obo):
    """An ontology representation of MMSigDB's gene set nomenclature."""

    ontology = bioversions_key = PREFIX
    typedefs = [has_participant]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise, force=force)


def get_obo(force: bool = False) -> Obo:
    """Get MSIG as Obo."""
    return MSigDBGetter(force=force)


_SPECIES = {
    "Homo sapiens": "9606",
    "Mus musculus": "10090",
    "Rattus norvegicus": "10116",
    "Macaca mulatta": "9544",
    "Danio rerio": "7955",
}

REACTOME_URL_PREFIX = "https://www.reactome.org/content/detail/"
GO_URL_PREFIX = "http://amigo.geneontology.org/amigo/term/GO:"
KEGG_URL_PREFIX = "http://www.genome.jp/kegg/pathway/hsa/"


def iter_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Get MSigDb terms."""
    xml_url = f"{BASE_URL}/{version}.Hs/msigdb_v{version}.Hs.xml"
    path = ensure_path(prefix=PREFIX, url=xml_url, version=version, force=force)
    tree = ElementTree.parse(path)

    for entry in tqdm(tree.getroot(), desc=f"{PREFIX} v{version}", unit_scale=True):
        attrib = dict(entry.attrib)
        tax_id = _SPECIES[attrib["ORGANISM"]]

        reference_id = attrib["PMID"].strip()
        if not reference_id:
            reference = None
        elif reference_id.startswith("GSE"):
            reference = Reference(prefix="gse", identifier=reference_id)
        else:
            reference = Reference(prefix="pubmed", identifier=reference_id)

        # NONE have the entry "HISTORICAL_NAME"
        # historical_name = thing.attrib['HISTORICAL_NAME']

        identifier = attrib["SYSTEMATIC_NAME"]
        name = attrib["STANDARD_NAME"]
        is_obsolete = attrib["CATEGORY_CODE"] == "ARCHIVED"

        term = Term(
            reference=Reference(prefix=PREFIX, identifier=identifier, name=name),
            definition=_get_definition(attrib),
            provenance=[] if reference is None else [reference],
            is_obsolete=is_obsolete,
        )
        for key in [
            "CATEGORY_CODE",
            "SUB_CATEGORY_CODE",
            "CONTRIBUTOR",
            "EXACT_SOURCE",
            "EXTERNAL_DETAILS_URL",
        ]:
            value = attrib[key].strip()
            if value:
                term.append_property(key.lower(), value)

        term.set_species(tax_id)

        contributor = attrib["CONTRIBUTOR"]
        external_id = attrib["EXACT_SOURCE"]
        external_details = attrib["EXTERNAL_DETAILS_URL"]
        if contributor == "WikiPathways":
            if not external_id:
                logger.warning(
                    "missing %s source: msigdb:%s (%s)", contributor, identifier, external_details
                )
            term.append_xref(Reference(prefix="wikipathways", identifier=external_id))
        elif contributor == "Reactome":
            if not external_id:
                logger.warning(
                    "missing %s source: msigdb:%s (%s)", contributor, identifier, external_details
                )
            term.append_xref(Reference(prefix="reactome", identifier=external_id))
        elif contributor == "Gene Ontology":
            if not external_id:
                external_id = external_details[len(GO_URL_PREFIX) :]
            if not external_id:
                logger.warning(
                    "missing %s source: msigdb:%s (%s)", contributor, identifier, external_details
                )
            term.append_xref(Reference(prefix="go", identifier=external_id))
        elif contributor == "KEGG":
            if not external_id:
                external_id = external_details[len(KEGG_URL_PREFIX) : len(".html")]
            if not external_id:
                logger.warning(
                    "missing %s source: msigdb:%s (%s)", contributor, identifier, external_details
                )
            term.append_xref(Reference(prefix="kegg.pathway", identifier=external_id))

        for ncbigene_id in attrib["MEMBERS_EZID"].strip().split(","):
            if ncbigene_id:
                term.append_relationship(
                    has_participant, Reference(prefix="ncbigene", identifier=ncbigene_id)
                )
        yield term


def _get_definition(attrib) -> Optional[str]:
    rv = attrib["DESCRIPTION_FULL"].strip() or attrib["DESCRIPTION_BRIEF"].strip() or None
    if rv is not None:
        return rv.replace("\d", "").replace("\s", "")  # noqa: W605
    return None


if __name__ == "__main__":
    MSigDBGetter.cli()



# -*- coding: utf-8 -*-

"""High-level API for alternative identifiers."""

import logging
from functools import lru_cache
from typing import List, Mapping, Optional

from .utils import get_version
from ..getters import get_ontology
from ..identifier_utils import normalize_curie, wrap_norm_prefix
from ..utils.cache import cached_multidict
from ..utils.path import prefix_cache_join

__all__ = [
    "get_id_to_alts",
    "get_alts_to_id",
    "get_primary_curie",
    "get_primary_identifier",
]

logger = logging.getLogger(__name__)

NO_ALTS = {
    "ncbigene",
}


@lru_cache()
@wrap_norm_prefix
def get_id_to_alts(prefix: str, force: bool = False) -> Mapping[str, List[str]]:
    """Get alternate identifiers."""
    if prefix in NO_ALTS:
        return {}

    version = get_version(prefix)
    path = prefix_cache_join(prefix, name="alt_ids.tsv", version=version)
    header = [f"{prefix}_id", "alt_id"]

    @cached_multidict(path=path, header=header, force=force)
    def _get_mapping() -> Mapping[str, List[str]]:
        if force:
            logger.info(f"[{prefix}] forcing reload for alts")
        else:
            logger.info("[%s] no cached alts found. getting from OBO loader", prefix)
        ontology = get_ontology(prefix, force=force, version=version)
        return ontology.get_id_alts_mapping()

    return _get_mapping()


@lru_cache()
@wrap_norm_prefix
def get_alts_to_id(prefix: str, force: bool = False) -> Mapping[str, str]:
    """Get alternative id to primary id mapping."""
    return {
        alt: primary
        for primary, alts in get_id_to_alts(prefix, force=force).items()
        for alt in alts
    }


def get_primary_curie(curie: str) -> Optional[str]:
    """Get the primary curie for an entity."""
    prefix, identifier = normalize_curie(curie)
    primary_identifier = get_primary_identifier(prefix, identifier)
    if primary_identifier is not None:
        return f"{prefix}:{primary_identifier}"
    return None


@wrap_norm_prefix
def get_primary_identifier(prefix: str, identifier: str) -> str:
    """Get the primary identifier for an entity.

    :param prefix: The name of the resource
    :param identifier: The identifier to look up
    :returns: the canonical identifier based on alt id lookup

    Returns the original identifier if there are no alts available or if there's no mapping.
    """
    if prefix in NO_ALTS:  # TODO later expand list to other namespaces with no alts
        return identifier

    alts_to_id = get_alts_to_id(prefix)
    if alts_to_id and identifier in alts_to_id:
        return alts_to_id[identifier]
    return identifier



# -*- coding: utf-8 -*-

"""Tools for iterating over things."""

import csv
import gzip
from typing import Iterable, List, Tuple, TypeVar

from more_itertools import peekable

__all__ = [
    "iterate_together",
    "iterate_gzips_together",
]

X = TypeVar("X")
Z = TypeVar("Z")
Y = TypeVar("Y")


def iterate_gzips_together(a_path, b_path) -> Iterable[Tuple[str, str, List[str]]]:
    """Iterate over two gzipped files together."""
    with gzip.open(a_path, mode="rt", errors="ignore") as a, gzip.open(b_path, mode="rt") as b:
        a = csv.reader(a, delimiter="\t", quoting=csv.QUOTE_MINIMAL)
        b = csv.reader(b, delimiter="\t", quoting=csv.QUOTE_MINIMAL)
        yield from iterate_together(a, b)


def iterate_together(
    a: Iterable[Tuple[X, Y]], b: Iterable[Tuple[X, Z]]
) -> Iterable[Tuple[X, Y, List[Z]]]:
    """Iterate over two sorted lists that have the same keys.

    The lists have to have the following invariants:

    - a is a one-to-one mapping.
    - b is a one-to-many mapping.
    - Both are indexed with the same numbers and in sorted order.
    - Each key in the index is present within both files
    """
    b_peekable = peekable(b)
    b_index, _ = b_peekable.peek()

    for a_index, a_value in a:
        zs = []
        while a_index == b_index:
            _, b_value = next(b_peekable)
            zs.append(b_value)
            b_index, _ = b_peekable.peek((_Done, _Done))
        yield a_index, a_value, zs


class _Done:
    pass



# -*- coding: utf-8 -*-

"""Convert DrugBank to OBO.

Run with ``python -m pyobo.sources.drugbank``
"""

import datetime
import itertools as itt
import logging
from functools import lru_cache
from typing import Any, Dict, Iterable, Mapping, Optional
from xml.etree import ElementTree

import pystow
from tqdm.auto import tqdm

from ..struct import Obo, Reference, Term
from ..struct.typedef import has_salt
from ..utils.cache import cached_pickle
from ..utils.path import prefix_directory_join

__all__ = [
    "DrugBankGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "drugbank"


class DrugBankGetter(Obo):
    """A getter for DrugBank."""

    ontology = bioversions_key = PREFIX
    typedefs = [has_salt]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise, force=force)


def get_obo(force: bool = False) -> Obo:
    """Get DrugBank as OBO."""
    return DrugBankGetter(force=force)


def iter_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Iterate over DrugBank terms in OBO."""
    for drug_info in iterate_drug_info(version, force=force):
        yield _make_term(drug_info)


def iterate_drug_info(version: str, force: bool = False) -> Iterable[Mapping[str, Any]]:
    """Iterate over DrugBank records."""

    @cached_pickle(
        prefix_directory_join(PREFIX, name="precompiled.pkl", version=version), force=force
    )
    def _inner():
        root = get_xml_root(version)
        rv = [_extract_drug_info(drug_xml) for drug_xml in tqdm(root, desc="Drugs")]
        return rv

    return _inner()


DRUG_XREF_SKIP = {
    "Wikipedia",
    "PDB",
    "Drugs Product Database (DPD)",  # TODO needs curating in metaregistry
    "RxCUI",  # TODO needs curating in metaregistry
    "GenBank",  # about protein
    "UniProtKB",  # about protein
    "Guide to Pharmacology",  # redundant of
}

DRUG_XREF_MAPPING = {
    "PubChem Compound": "pubchem.compound",
    "PubChem Substance": "pubchem.substance",
    "ChEBI": "chebi",
    "KEGG Drug": "kegg.drug",
    "KEGG Compound": "kegg.compound",
    "ChemSpider": "chemspider",
    "ChEMBL": "chembl.compound",
    "ZINC": "zinc",
    "BindingDB": "bindingdb",
    "PharmGKB": "pharmgkb.drug",
    "Therapeutic Targets Database": "ttd.drug",
    "IUPHAR": "iuphar.ligand",
}


def _make_term(drug_info: Mapping[str, Any]) -> Term:
    term = Term.from_triple(
        prefix=PREFIX,
        identifier=drug_info["drugbank_id"],
        name=drug_info["name"],
    )
    definition = drug_info.get("description")
    if definition:
        definition = definition.strip('"').replace("\t", " ").replace("  ", " ")
        term.definition = definition

    for alias in drug_info["aliases"]:
        term.append_synonym(alias)

    for xref in drug_info["xrefs"]:
        xref_prefix, xref_identifier = xref["resource"], xref["identifier"]
        if xref_prefix in DRUG_XREF_SKIP:
            continue
        xref_prefix_norm = DRUG_XREF_MAPPING.get(xref_prefix)
        if xref_prefix_norm is None:
            logger.warning("unhandled xref: %s:%s", xref_prefix, xref_identifier)
            continue
        term.append_xref(Reference(prefix=xref_prefix_norm, identifier=xref_identifier))

    for xref_prefix in ["cas", "inchikey"]:
        identifier = drug_info.get(xref_prefix)
        if identifier:
            term.append_xref(Reference(prefix=xref_prefix, identifier=identifier))

    for prop in ["smiles", "inchi"]:
        identifier = drug_info.get(xref_prefix)
        if identifier:
            term.append_property(prop, identifier)

    for salt in drug_info.get("salts", []):
        term.append_relationship(
            has_salt,
            Reference(
                prefix="drugbank.salt",
                identifier=salt["identifier"],
                name=salt["name"],
            ),
        )

    return term


@lru_cache()
def get_xml_root(version: Optional[str] = None) -> ElementTree.Element:
    """Get the DrugBank XML parser root.

    Takes between 35-60 seconds.
    """
    from drugbank_downloader import parse_drugbank

    element = parse_drugbank(
        version=version,
        username=pystow.get_config("pyobo", "drugbank_username"),
        password=pystow.get_config("pyobo", "drugbank_password"),
    )
    return element.getroot()


ns = "{http://www.drugbank.ca}"
inchikey_template = f"{ns}calculated-properties/{ns}property[{ns}kind='InChIKey']/{ns}value"
inchi_template = f"{ns}calculated-properties/{ns}property[{ns}kind='InChI']/{ns}value"
smiles_template = f"{ns}calculated-properties/{ns}property[{ns}kind='SMILES']/{ns}value"


def _extract_drug_info(drug_xml: ElementTree.Element) -> Mapping[str, Any]:
    """Extract information from an XML element representing a drug."""
    # assert drug_xml.tag == f'{ns}drug'
    row: Dict[str, Any] = {
        "type": drug_xml.get("type"),
        "drugbank_id": drug_xml.findtext(f"{ns}drugbank-id[@primary='true']"),
        "cas": drug_xml.findtext(f"{ns}cas-number"),
        "name": drug_xml.findtext(f"{ns}name"),
        "groups": [group.text for group in drug_xml.findall(f"{ns}groups/{ns}group")],
        "atc_codes": [code.get("code") for code in drug_xml.findall(f"{ns}atc-codes/{ns}atc-code")],
        "categories": [
            {
                "name": x.findtext(f"{ns}category"),
                "mesh_id": x.findtext(f"{ns}mesh-id"),
            }
            for x in drug_xml.findall(f"{ns}categories/{ns}category")
        ],
        "patents": list(_get_patents(drug_xml)),
        "salts": [
            {
                "identifier": x.findtext(f"{ns}drugbank-id"),
                "name": x.findtext(f"{ns}name"),
                "unii": x.findtext(f"{ns}unii"),
                "cas": x.findtext(f"{ns}cas-number"),
                "inchikey": x.findtext(f"{ns}inchikey"),
            }
            for x in drug_xml.findall(f"{ns}salts/{ns}salt")
        ],
        "xrefs": [
            {
                "resource": x.findtext(f"{ns}resource"),
                "identifier": x.findtext(f"{ns}identifier"),
            }
            for x in drug_xml.findall(f"{ns}external-identifiers/{ns}external-identifier")
        ],
        "inchi": drug_xml.findtext(inchi_template),
        "inchikey": drug_xml.findtext(inchikey_template),
        "smiles": drug_xml.findtext(smiles_template),
    }

    description = drug_xml.findtext(f"{ns}description")
    if description:
        row["description"] = description.replace("\r", "").replace("\n", "\\n")

    # Add drug aliases
    aliases = {
        elem.text.strip()
        for elem in itt.chain(
            drug_xml.findall(f"{ns}international-brands/{ns}international-brand"),
            drug_xml.findall(f"{ns}synonyms/{ns}synonym[@language='English']"),
            drug_xml.findall(f"{ns}international-brands/{ns}international-brand"),
            drug_xml.findall(f"{ns}products/{ns}product/{ns}name"),
        )
        if elem.text and elem.text.strip()
    }
    aliases.add(row["name"])
    row["aliases"] = aliases

    row["protein_interactions"] = []
    row["protein_group_interactions"] = []

    for category, protein in _iterate_protein_stuff(drug_xml):
        target_row = _extract_protein_info(category, protein)
        if not target_row:
            continue
        row["protein_interactions"].append(target_row)

    return row


def _get_patents(drug_element):
    for patent_element in drug_element.findall(f"{ns}patents/{ns}patent"):
        rv = {
            "patent_id": patent_element.findtext(f"{ns}number"),
            "country": patent_element.findtext(f"{ns}country"),
            "pediatric_extension": patent_element.findtext(f"{ns}pediatric-extension") != "false",
        }
        approved = patent_element.findtext(f"{ns}approved")
        if approved is not None:
            rv["approved"] = datetime.datetime.strptime(approved, "%Y-%m-%d")
        expires = patent_element.findtext(f"{ns}expires")
        if expires:
            rv["expires"] = datetime.datetime.strptime(expires, "%Y-%m-%d")
        yield rv


_categories = ["target", "enzyme", "carrier", "transporter"]


def _iterate_protein_stuff(drug_xml):
    for category in _categories:
        proteins = drug_xml.findall(f"{ns}{category}s/{ns}{category}")
        for protein in proteins:
            yield category, protein


def _extract_protein_info(category, protein):
    # FIXME Differentiate between proteins and protein groups/complexes
    polypeptides = protein.findall(f"{ns}polypeptide")

    if len(polypeptides) == 0:
        protein_type = "none"
    elif len(polypeptides) == 1:
        protein_type = "single"
    else:
        protein_type = "group"

    row = {
        "target": {
            "type": protein_type,
            "category": category,
            "known_action": protein.findtext(f"{ns}known-action"),
            "name": protein.findtext(f"{ns}name"),
            "actions": [action.text for action in protein.findall(f"{ns}actions/{ns}action")],
            "articles": [
                pubmed_element.text
                for pubmed_element in protein.findall(
                    f"{ns}references/{ns}articles/{ns}article/{ns}pubmed-id"
                )
                if pubmed_element.text
            ],
        },
        "polypeptides": list(_iter_polypeptides(polypeptides)),
    }
    return row


def _iter_polypeptides(polypeptides) -> Iterable[Mapping[str, Any]]:
    for polypeptide in polypeptides:
        name = polypeptide.findtext(f"{ns}name")

        uniprot_id = polypeptide.findtext(
            f"{ns}external-identifiers/{ns}external-identifier[{ns}resource='UniProtKB']/{ns}identifier",
        )
        uniprot_accession = polypeptide.findtext(
            f"{ns}external-identifiers/{ns}external-identifier[{ns}resource='UniProt Accession']/{ns}identifier",
        )
        organism = polypeptide.findtext(f"{ns}organism")
        taxonomy_id = polypeptide.find(f"{ns}organism").attrib["ncbi-taxonomy-id"]

        yv = {
            "name": name,
            "uniprot_id": uniprot_id,
            "uniprot_accession": uniprot_accession,
            "organism": organism,
            "taxonomy": taxonomy_id,
        }

        hgnc_id = polypeptide.findtext(
            f"{ns}external-identifiers/{ns}external-identifier"
            f"[{ns}resource='HUGO Gene Nomenclature Committee (HGNC)']/{ns}identifier",
        )
        if hgnc_id is not None:
            hgnc_id = hgnc_id[len("HGNC:") :]
            yv["hgnc_id"] = hgnc_id
            yv["hgnc_symbol"] = polypeptide.findtext(f"{ns}gene-name")

        yield yv


if __name__ == "__main__":
    DrugBankGetter.cli()



# -*- coding: utf-8 -*-

"""Converter for Reactome."""

import logging
from collections import defaultdict
from functools import lru_cache
from typing import Iterable, Mapping, Set

import pandas as pd
from tqdm.auto import tqdm

from ..api import get_id_multirelations_mapping
from ..constants import SPECIES_REMAPPING
from ..resources.ncbitaxon import get_ncbitaxon_id
from ..struct import Obo, Reference, Term, from_species, has_participant
from ..utils.io import multidict
from ..utils.path import ensure_df

__all__ = [
    "ReactomeGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "reactome"


# TODO alt ids https://reactome.org/download/current/reactome_stable_ids.txt


class ReactomeGetter(Obo):
    """An ontology representation of the Reactome pathway database."""

    ontology = bioversions_key = PREFIX
    typedefs = [from_species, has_participant]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise, force=force)


def get_obo(force: bool = False) -> Obo:
    """Get Reactome OBO."""
    return ReactomeGetter(force=force)


def ensure_participant_df(version: str, force: bool = False) -> pd.DataFrame:
    """Get the pathway uniprot participant dataframe."""
    uniprot_pathway_url = f"https://reactome.org/download/{version}/UniProt2Reactome_All_Levels.txt"
    return ensure_df(
        PREFIX, url=uniprot_pathway_url, header=None, usecols=[0, 1], version=version, force=force
    )


def iter_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Iterate Reactome terms."""
    provenance_url = f"https://reactome.org/download/{version}/ReactionPMIDS.txt"
    provenance_df = ensure_df(PREFIX, url=provenance_url, header=None, version=version, force=force)
    provenance_d = multidict(provenance_df.values)

    pathway_names_url = f"https://reactome.org/download/{version}/ReactomePathways.txt"
    df = ensure_df(
        PREFIX,
        url=pathway_names_url,
        header=None,
        names=["reactome_id", "name", "species"],
        version=version,
        force=force,
    )
    df["species"] = df["species"].map(lambda x: SPECIES_REMAPPING.get(x) or x)
    df["taxonomy_id"] = df["species"].map(get_ncbitaxon_id)

    terms = {}
    it = tqdm(df.values, total=len(df.index), desc=f"mapping {PREFIX}")
    for reactome_id, name, species_name, taxonomy_id in it:
        terms[reactome_id] = term = Term(
            reference=Reference(prefix=PREFIX, identifier=reactome_id, name=name),
            provenance=[
                Reference(prefix="pubmed", identifier=pubmed_id)
                for pubmed_id in provenance_d.get(reactome_id, [])
            ],
        )
        if not taxonomy_id or pd.isna(taxonomy_id):
            raise ValueError(f"unmapped species: {species_name}")

        term.set_species(identifier=taxonomy_id, name=species_name)

    pathways_hierarchy_url = f"https://reactome.org/download/{version}/ReactomePathwaysRelation.txt"
    hierarchy_df = ensure_df(
        PREFIX, url=pathways_hierarchy_url, header=None, version=version, force=force
    )
    for parent_id, child_id in hierarchy_df.values:
        terms[child_id].append_parent(terms[parent_id])

    uniprot_pathway_df = ensure_participant_df(version=version, force=force)
    for uniprot_id, reactome_id in tqdm(uniprot_pathway_df.values, total=len(uniprot_pathway_df)):
        terms[reactome_id].append_relationship(
            has_participant, Reference(prefix="uniprot", identifier=uniprot_id)
        )

    chebi_pathway_url = f"https://reactome.org/download/{version}/ChEBI2Reactome_All_Levels.txt"
    chebi_pathway_df = ensure_df(
        PREFIX,
        url=chebi_pathway_url,
        header=None,
        usecols=[0, 1],
        version=version,
        force=force,
    )
    for chebi_id, reactome_id in tqdm(chebi_pathway_df.values, total=len(chebi_pathway_df)):
        terms[reactome_id].append_relationship(
            has_participant, Reference(prefix="chebi", identifier=chebi_id)
        )

    # ncbi_pathway_url = f'https://reactome.org/download/{version}/NCBI2Reactome_All_Levels.txt'
    # ncbi_pathway_df = ensure_df(PREFIX, url=ncbi_pathway_url, header=None, usecols=[0, 1], version=version)
    # for ncbigene_id, reactome_id in tqdm(ncbi_pathway_df.values, total=len(ncbi_pathway_df)):
    #     terms[reactome_id].append_relationship(has_part, Reference('ncbigene', ncbigene_id))

    yield from terms.values()


@lru_cache(maxsize=1)
def get_protein_to_pathways() -> Mapping[str, Set[str]]:
    """Get a mapping from proteins to the pathways they're in."""
    protein_to_pathways = defaultdict(set)
    x = get_id_multirelations_mapping("reactome", has_participant)
    for reactome_id, proteins in x.items():
        for protein in proteins:
            if protein.prefix != "uniprot":
                continue
            protein_to_pathways[protein.identifier].add(reactome_id)
    return dict(protein_to_pathways)


if __name__ == "__main__":
    get_obo().write_default()



# -*- coding: utf-8 -*-

"""API utilities for KEGG."""

import urllib.error
from dataclasses import dataclass
from typing import Mapping, Optional

from pyobo import Reference, Term, ensure_path
from pyobo.struct import from_species
from pyobo.utils.path import ensure_df, prefix_directory_join

KEGG_GENES_PREFIX = "kegg.genes"
KEGG_GENOME_PREFIX = "kegg.genome"
KEGG_PATHWAY_PREFIX = "kegg.pathway"

BASE = "http://rest.kegg.jp"
SKIP = {
    "T03333",
    "T03334",
    "T03356",
    "T03357",
    "T03358",
    "T03359",
}


@dataclass
class KEGGGenome:
    """A data structure for a parsed line of the KEGG Genomes list."""

    identifier: str
    name: str
    code: Optional[str]
    long_code: Optional[str]
    taxonomy_id: Optional[str]

    def annotate_term(self, term: Term) -> None:
        """Annotate the term with the species represented by this object."""
        term.append_relationship(
            from_species,
            self.get_reference(),
        )
        if self.taxonomy_id is not None:
            term.set_species(self.taxonomy_id)

    def get_reference(self) -> Reference:
        """Get the reference for this genome."""
        return Reference(
            prefix="kegg.genome",
            identifier=self.identifier,
            name=self.name,
        )


def ensure_list_genomes(version: str) -> str:
    """Ensure the KEGG Genome file is downloaded."""
    return ensure_path(
        KEGG_GENOME_PREFIX,
        url=f"{BASE}/list/genome",
        name="genome.tsv",
        version=version,
    )


def ensure_list_pathways(version: str) -> Mapping[str, str]:
    """Ensure the KEGG Map (non species specific)."""
    rv = ensure_df(
        KEGG_PATHWAY_PREFIX,
        url=f"{BASE}/list/pathway",
        name="pathway.tsv",
        version=version,
    )
    return {k[len("path:") :]: v for k, v in rv.values}


"""GENOME SPECIFIC"""


def ensure_list_genome(kegg_genome_id: str, *, version: str) -> str:
    """Get the list of genes for the given organism."""
    return ensure_path(
        KEGG_GENES_PREFIX,
        "genes",
        url=f"{BASE}/list/{kegg_genome_id}",
        name=f"{kegg_genome_id}.tsv",
        version=version,
    )


def ensure_conv_genome_uniprot(
    kegg_genome_id: str, *, version: str, error_on_missing: bool = False
) -> Optional[str]:
    """Get the KEGG-UniProt protein map for the given organism."""
    return _ensure_conv_genome_helper(
        kegg_genome_id, "uniprot", version=version, error_on_missing=error_on_missing
    )


def ensure_conv_genome_ncbigene(
    kegg_genome_id: str, *, version: str, error_on_missing: bool = False
) -> Optional[str]:
    """Get the KEGG-NCBIGENE protein map for the given organism."""
    return _ensure_conv_genome_helper(
        kegg_genome_id, "ncbi-geneid", version=version, error_on_missing=error_on_missing
    )


def _ensure_conv_genome_helper(
    kegg_genome_id: str,
    target_database: str,
    *,
    version: str,
    error_on_missing: bool = False,
) -> Optional[str]:
    """Get the KEGG-external protein map for the given organism/database."""
    name = f"{kegg_genome_id}.tsv"
    try:
        rv = ensure_path(
            KEGG_GENES_PREFIX,
            f"conv_{target_database}",
            url=f"{BASE}/conv/{target_database}/{kegg_genome_id}",
            name=name,
            error_on_missing=error_on_missing,
            version=version,
        )
    except urllib.error.HTTPError:
        path_rv = prefix_directory_join(
            KEGG_GENES_PREFIX,
            f"conv_{target_database}",
            name=name,
            version=version,
        )
        with path_rv.open("w") as file:
            print(file=file)  # noqa: T201
        return path_rv.as_posix()
    except FileNotFoundError:
        return None
    else:
        return rv


def ensure_link_pathway_genome(
    kegg_genome_id: str, *, version: str, error_on_missing: bool = False
) -> str:
    """Get the protein-pathway links for the given organism.

    :raises: FileNotFoundError
    """
    return ensure_path(
        KEGG_PATHWAY_PREFIX,
        "link_pathway",
        url=f"{BASE}/link/pathway/{kegg_genome_id}",
        name=f"{kegg_genome_id}.tsv",
        error_on_missing=error_on_missing,
        version=version,
    )


def ensure_list_pathway_genome(
    kegg_genome_id: str, *, version: str, error_on_missing: bool = False
) -> str:
    """Get the list of pathways for the given organism.

    :raises: FileNotFoundError
    """
    return ensure_path(
        KEGG_PATHWAY_PREFIX,
        "pathways",
        url=f"{BASE}/list/pathway/{kegg_genome_id}",
        name=f"{kegg_genome_id}.tsv",
        error_on_missing=error_on_missing,
        version=version,
    )



# -*- coding: utf-8 -*-

"""Tools for canonicalizing a CURIE based on a priority list."""

from dataclasses import dataclass, field
from functools import lru_cache
from typing import Iterable, List, Mapping, Optional, Set, Tuple

import networkx as nx
import pandas as pd
from more_itertools import pairwise
from tqdm.auto import tqdm

from .priority import DEFAULT_PRIORITY_LIST
from .xrefs_pipeline import get_graph_from_xref_df
from .. import resource_utils
from ..utils.io import get_reader, get_writer

__all__ = [
    "Canonicalizer",
    "all_shortest_paths",
    "single_source_shortest_path",
    "get_equivalent",
    "get_priority_curie",
    "remap_file_stream",
]


@dataclass
class Canonicalizer:
    """Wraps a graph and priority list to allow getting the best identifier."""

    #: A graph from :func:`get_graph_from_xref_df`
    graph: nx.Graph

    #: A list of prefixes. The ones with the lower index are higher priority
    priority: Optional[List[str]] = None

    #: Longest length paths allowed
    cutoff: int = 5

    _priority: Mapping[str, int] = field(init=False)

    def __post_init__(self):
        """Initialize the priority map based on the priority list."""
        if self.priority is None:
            self.priority = DEFAULT_PRIORITY_LIST
        self._priority = {entry: len(self.priority) - i for i, entry in enumerate(self.priority)}

    def _key(self, curie: str) -> Optional[int]:
        prefix = self.graph.nodes[curie]["prefix"]
        return self._priority.get(prefix)

    def _get_priority_dict(self, curie: str) -> Mapping[str, int]:
        return dict(self._iterate_priority_targets(curie))

    def _iterate_priority_targets(self, curie: str) -> Iterable[Tuple[str, int]]:
        for target in nx.single_source_shortest_path(self.graph, curie, cutoff=self.cutoff):
            priority = self._key(target)
            if priority is not None:
                yield target, priority
            elif target == curie:
                yield target, 0
            else:
                yield target, -1

    def canonicalize(self, curie: str) -> str:
        """Get the best CURIE from the given CURIE."""
        if curie not in self.graph:
            return curie
        priority_dict = self._get_priority_dict(curie)
        return max(priority_dict, key=priority_dict.get)  # type:ignore

    @classmethod
    def get_default(cls, priority: Optional[Iterable[str]] = None) -> "Canonicalizer":
        """Get the default canonicalizer."""
        if priority is not None:
            priority = tuple(priority)
        return cls._get_default_helper(priority=priority)

    @classmethod
    @lru_cache()
    def _get_default_helper(cls, priority: Optional[Tuple[str, ...]] = None) -> "Canonicalizer":
        """Help get the default canonicalizer."""
        graph = cls._get_default_graph()
        return cls(graph=graph, priority=list(priority) if priority else None)

    @staticmethod
    @lru_cache()
    def _get_default_graph() -> nx.Graph:
        df = resource_utils.ensure_inspector_javert_df()
        graph = get_graph_from_xref_df(df)
        return graph

    def iterate_flat_mapping(self, use_tqdm: bool = True) -> Iterable[Tuple[str, str]]:
        """Iterate over the canonical mapping from all nodes to their canonical CURIEs."""
        nodes = self.graph.nodes()
        if use_tqdm:
            nodes = tqdm(
                nodes,
                total=self.graph.number_of_nodes(),
                desc="building flat mapping",
                unit_scale=True,
                unit="CURIE",
            )
        for node in nodes:
            yield node, self.canonicalize(node)

    def get_flat_mapping(self, use_tqdm: bool = True) -> Mapping[str, str]:
        """Get a canonical mapping from all nodes to their canonical CURIEs."""
        return dict(self.iterate_flat_mapping(use_tqdm=use_tqdm))

    def single_source_shortest_path(
        self,
        curie: str,
        cutoff: Optional[int] = None,
    ) -> Optional[Mapping[str, List[Mapping[str, str]]]]:
        """Get all shortest paths between given entity and its equivalent entities."""
        return single_source_shortest_path(graph=self.graph, curie=curie, cutoff=cutoff)

    def all_shortest_paths(
        self, source_curie: str, target_curie: str
    ) -> List[List[Mapping[str, str]]]:
        """Get all shortest paths between the two entities."""
        return all_shortest_paths(
            graph=self.graph, source_curie=source_curie, target_curie=target_curie
        )

    @classmethod
    def from_df(cls, df: pd.DataFrame) -> "Canonicalizer":
        """Instantiate from a dataframe."""
        return cls(graph=get_graph_from_xref_df(df))


def all_shortest_paths(
    graph: nx.Graph, source_curie: str, target_curie: str
) -> List[List[Mapping[str, str]]]:
    """Get all shortest paths between the two CURIEs."""
    _paths = nx.all_shortest_paths(graph, source=source_curie, target=target_curie)
    return [
        [dict(source=s, target=t, provenance=graph[s][t]["source"]) for s, t in pairwise(_path)]
        for _path in _paths
    ]


def single_source_shortest_path(
    graph: nx.Graph,
    curie: str,
    cutoff: Optional[int] = None,
) -> Optional[Mapping[str, List[Mapping[str, str]]]]:
    """Get the shortest path from the CURIE to all elements of its equivalence class.

    Things that didn't work:

    Unresponsive
    ------------
    .. code-block:: python

        for curies in tqdm(nx.connected_components(graph), desc='filling connected components', unit_scale=True):
            for c1, c2 in itt.combinations(curies, r=2):
                if not graph.has_edge(c1, c2):
                    graph.add_edge(c1, c2, inferred=True)

    Way too slow
    ------------
    .. code-block:: python

        for curie in tqdm(graph, total=graph.number_of_nodes(), desc='mapping connected components', unit_scale=True):
            for incident_curie in nx.node_connected_component(graph, curie):
                if not graph.has_edge(curie, incident_curie):
                    graph.add_edge(curie, incident_curie, inferred=True)

    Also consider the condensation of the graph:
    https://networkx.github.io/documentation/stable/reference/algorithms/generated/networkx.algorithms.components.condensation.html#networkx.algorithms.components.condensation
    """
    if curie not in graph:
        return None
    rv = nx.single_source_shortest_path(graph, curie, cutoff=cutoff)
    return {
        k: [dict(source=s, target=t, provenance=graph[s][t]["provenance"]) for s, t in pairwise(v)]
        for k, v in rv.items()
        if k != curie  # don't map to self
    }


def get_equivalent(curie: str, cutoff: Optional[int] = None) -> Set[str]:
    """Get equivalent CURIEs."""
    canonicalizer = Canonicalizer.get_default()
    r = canonicalizer.single_source_shortest_path(curie=curie, cutoff=cutoff)
    return set(r or [])


def get_priority_curie(curie: str) -> str:
    """Get the priority CURIE mapped to the best namespace."""
    canonicalizer = Canonicalizer.get_default()
    return canonicalizer.canonicalize(curie)


def remap_file_stream(file_in, file_out, column: int, sep="\t") -> None:
    """Remap a file."""
    reader = get_reader(file_in, sep=sep)
    writer = get_writer(file_out, sep=sep)
    for row in reader:
        row[column] = get_priority_curie(row[column])
        writer.writerow(row)



# -*- coding: utf-8 -*-

"""Selventa diseases.

.. seealso:: https://github.com/pyobo/pyobo/issues/26
"""

from typing import Iterable

import pandas as pd

from pyobo import Obo, Term
from pyobo.utils.path import ensure_df

__all__ = [
    "SDISGetter",
]

PREFIX = "sdis"
URL = "https://raw.githubusercontent.com/OpenBEL/resource-generator/master/datasets/selventa-legacy-diseases.txt"


class SDISGetter(Obo):
    """An ontology representation of the Selventa disease nomenclature."""

    ontology = PREFIX
    static_version = "1.0.0"

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(force=force)


def get_obo(*, force: bool = False) -> Obo:
    """Get Selventa Diseases as OBO."""
    return SDISGetter(force=force)


def iter_terms(force: bool = False) -> Iterable[Term]:
    """Iterate over selventa disease terms."""
    df = ensure_df(PREFIX, url=URL, skiprows=9, force=force)

    for identifier, label, synonyms, xrefs in df[["ID", "LABEL", "SYNONYMS", "XREF"]].values:
        term = Term.from_triple(PREFIX, identifier, label)
        for synonym in synonyms.split("|") if pd.notna(synonyms) else []:
            term.append_synonym(synonym)
        for xref in xrefs.split("|") if pd.notna(xrefs) else []:
            term.append_xref(xref)
        yield term


if __name__ == "__main__":
    get_obo().write_default(write_obo=True, force=True)



# -*- coding: utf-8 -*-

"""Converter for miRBase Mature."""

from typing import Iterable

import pandas as pd
from tqdm.auto import tqdm

from .mirbase_constants import get_mature_df
from ..struct import Obo, Reference, Synonym, Term

__all__ = [
    "MiRBaseMatureGetter",
]

PREFIX = "mirbase.mature"


class MiRBaseMatureGetter(Obo):
    """An ontology representation of miRBase's mature miRNA nomenclature."""

    ontology = PREFIX
    bioversions_key = "mirbase"

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise, force=force)


def get_obo(force: bool = False) -> Obo:
    """Get miRBase mature as OBO."""
    return MiRBaseMatureGetter(force=force)


def iter_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Get miRBase mature terms."""
    df = get_mature_df(version, force=force)
    for _, name, previous_name, mirbase_mature_id in tqdm(
        df.values, total=len(df.index), unit_scale=True
    ):
        yield Term(
            reference=Reference(
                prefix=PREFIX, identifier=mirbase_mature_id, name=name if pd.notna(name) else None
            ),
            synonyms=[
                Synonym(name=previous_name),
            ],
        )


if __name__ == "__main__":
    MiRBaseMatureGetter.cli()



# -*- coding: utf-8 -*-

"""Sources of xrefs not from OBO."""

import logging
from functools import lru_cache
from typing import Callable, Iterable, Mapping, Optional

import pandas as pd
from pkg_resources import iter_entry_points
from tqdm.auto import tqdm

__all__ = [
    "iter_xref_plugins",
    "has_xref_plugin",
    "run_xref_plugin",
    "iter_xref_plugins",
]

logger = logging.getLogger(__name__)


@lru_cache()
def _get_xref_plugins() -> Mapping[str, Callable[[], pd.DataFrame]]:
    return {entry.name: entry.load() for entry in iter_entry_points(group="pyobo.xrefs")}


def has_xref_plugin(prefix: str) -> bool:
    """Check if there's a plugin for converting the prefix."""
    return prefix in _get_xref_plugins()


def run_xref_plugin(prefix: str) -> pd.DataFrame:
    """Get a converted PyOBO source."""
    rv = _get_xref_plugins()[prefix]()

    if isinstance(rv, pd.DataFrame):
        return rv

    logger.warning("can not load %s since it yields many dataframes", prefix)


def iter_xref_plugins(
    use_tqdm: bool = True, skip_below: Optional[str] = None
) -> Iterable[pd.DataFrame]:
    """Get all modules in the PyOBO sources."""
    it = tqdm(sorted(_get_xref_plugins().items()), desc="Mapping Plugins", disable=not use_tqdm)
    for prefix, get_df in it:
        if skip_below and prefix < skip_below:
            continue
        it.set_postfix({"prefix": prefix})
        rv = get_df()
        if isinstance(rv, pd.DataFrame):
            yield rv
        elif isinstance(rv, Iterable):
            yield from rv
        else:
            raise TypeError



# -*- coding: utf-8 -*-

"""Converter for MGI."""

import logging
from collections import defaultdict
from typing import Iterable

import pandas as pd
from tqdm.auto import tqdm

from ..struct import (
    Obo,
    Reference,
    Synonym,
    Term,
    from_species,
    has_gene_product,
    transcribes_to,
)
from ..utils.path import ensure_df

__all__ = [
    "MGIGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "mgi"
MARKERS_URL = "http://www.informatics.jax.org/downloads/reports/MRK_List2.rpt"
ENTREZ_XREFS_URL = "http://www.informatics.jax.org/downloads/reports/MGI_EntrezGene.rpt"
ENSEMBL_XREFS_URL = "http://www.informatics.jax.org/downloads/reports/MRK_ENSEMBL.rpt"


class MGIGetter(Obo):
    """An ontology representation of MGI's mouse gene nomenclature."""

    ontology = PREFIX
    dynamic_version = True
    typedefs = [from_species, has_gene_product, transcribes_to]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return get_terms(force=force)


def get_obo(force: bool = False) -> Obo:
    """Get MGI as OBO."""
    return MGIGetter(force=force)


COLUMNS = ["MGI Accession ID", "Marker Symbol", "Marker Name"]


def get_ensembl_df(force: bool = False) -> pd.DataFrame:
    """Get the Ensembl mappings dataframe."""
    return ensure_df(
        PREFIX,
        url=ENSEMBL_XREFS_URL,
        sep="\t",
        names=[
            "mgi_id",
            "symbol",
            "name",
            "position",
            "chromosome",
            "ensembl_accession_id",
            "ensembl_transcript_id",
            "ensembl_protein_id",
            "feature_type",
            "start",
            "end",
            "strand",
            "biotypes",
        ],
        force=force,
        dtype=str,
    )


def get_entrez_df(
    force: bool = False,
) -> pd.DataFrame:
    """Get the Entrez mappings dataframe."""
    return ensure_df(
        PREFIX,
        url=ENTREZ_XREFS_URL,
        sep="\t",
        names=[
            "mgi_id",
            "symbol",
            "status",
            "name",
            "position",
            "chromosome",
            "Type",
            "Secondary Accession IDs",
            "entrez_id",
            "synonyms",
            "Features",
            "start",
            "end",
            "strand",
            "biotypes",
        ],
        dtype=str,
        force=force,
    )


def get_terms(force: bool = False) -> Iterable[Term]:
    """Get the MGI terms."""
    df = ensure_df(PREFIX, url=MARKERS_URL, sep="\t", force=force)

    entrez_df = get_entrez_df(force=force)
    mgi_to_entrez_id, mgi_to_synonyms = {}, {}
    for mgi_curie, synonyms, entrez_id in entrez_df[["mgi_id", "synonyms", "entrez_id"]].values:
        mgi_id = mgi_curie[len("MGI:") :]
        if synonyms and pd.notna(synonyms):
            mgi_to_synonyms[mgi_id] = synonyms.split("|")
        if entrez_id and pd.notna(entrez_id):
            mgi_to_entrez_id[mgi_id] = entrez_id

    ensembl_df = get_ensembl_df(force=force)
    # ensembl_df.to_csv('test.tsv', sep='\t', index=False)
    mgi_to_ensemble_accession_ids = defaultdict(list)
    mgi_to_ensemble_transcript_ids = defaultdict(list)
    mgi_to_ensemble_protein_ids = defaultdict(list)
    ensembl_columns = [
        "mgi_id",
        "ensembl_accession_id",
        "ensembl_transcript_id",
        "ensembl_protein_id",
    ]
    ensembl_it = ensembl_df[ensembl_columns].values
    for (
        mgi_curie,
        ensemble_accession_id,
        ensemble_transcript_ids,
        ensemble_protein_ids,
    ) in ensembl_it:
        mgi_id = mgi_curie[len("MGI:") :]
        if ensemble_accession_id and pd.notna(ensemble_accession_id):
            mgi_to_ensemble_accession_ids[mgi_id].append(ensemble_accession_id)
        if ensemble_transcript_ids and pd.notna(ensemble_transcript_ids):
            for ensemble_transcript_id in ensemble_transcript_ids.split():
                mgi_to_ensemble_transcript_ids[mgi_id].append(ensemble_transcript_id)
        if ensemble_protein_ids and pd.notna(ensemble_protein_ids):
            for ensemble_protein_id in ensemble_protein_ids.split():
                mgi_to_ensemble_protein_ids[mgi_id].append(ensemble_protein_id)

    for mgi_curie, name, definition in tqdm(
        df[COLUMNS].values, total=len(df.index), desc=f"Mapping {PREFIX}", unit_scale=True
    ):
        identifier = mgi_curie[len("MGI:") :]
        term = Term(
            reference=Reference(prefix=PREFIX, identifier=identifier, name=name),
            definition=definition,
        )
        if identifier in mgi_to_synonyms:
            for synonym in mgi_to_synonyms[identifier]:
                term.append_synonym(Synonym(name=synonym))
        if identifier in mgi_to_entrez_id:
            term.append_xref(Reference(prefix="ncbigene", identifier=mgi_to_entrez_id[identifier]))
        for ensembl_id in mgi_to_ensemble_accession_ids[identifier]:
            term.append_xref(Reference(prefix="ensembl", identifier=ensembl_id))
        for ensembl_id in mgi_to_ensemble_transcript_ids[identifier]:
            term.append_relationship(
                transcribes_to, Reference(prefix="ensembl", identifier=ensembl_id)
            )
        for ensembl_id in mgi_to_ensemble_protein_ids[identifier]:
            term.append_relationship(
                has_gene_product, Reference(prefix="ensembl", identifier=ensembl_id)
            )
        term.set_species(identifier="10090", name="Mus musculus")
        yield term


if __name__ == "__main__":
    get_obo(force=True).write_default(write_obo=True, write_obograph=True, use_tqdm=True)



# -*- coding: utf-8 -*-

"""Converter for CGNC."""

import logging
from typing import Iterable

import pandas as pd

from pyobo.struct import Obo, Reference, Term, from_species
from pyobo.utils.path import ensure_df

__all__ = [
    "CGNCGetter",
]

PREFIX = "cgnc"
URL = "http://birdgenenames.org/cgnc/downloads.jsp?file=standard"

logger = logging.getLogger(__name__)


class CGNCGetter(Obo):
    """An ontology representation of the Chicken Genome Nomenclature Consortium's gene nomenclature."""

    ontology = PREFIX
    dynamic_version = True
    typedefs = [from_species]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return get_terms(force=force)


def get_obo(force: bool = False) -> Obo:
    """Get CGNC as OBO."""
    return CGNCGetter(force=force)


HEADER = [
    "cgnc_id",
    "ncbigene_id",
    "ensembl_id",
    "name",
    "synonym_1",
    "synonym_2",
    "curation status",
    "last edit date",
]


def get_terms(force: bool = False) -> Iterable[Term]:
    """Get CGNC terms."""
    df = ensure_df(PREFIX, url=URL, name=f"{PREFIX}.tsv", force=force, header=0, names=HEADER)
    for i, (cgnc_id, entrez_id, ensembl_id, name, synonym_1, synoynm_2, _, _) in enumerate(
        df.values
    ):
        if pd.isna(cgnc_id):
            logger.warning(f"row {i} CGNC ID is none")
            continue

        try:
            int(cgnc_id)
        except ValueError:
            logger.warning(f"row {i} CGNC ID is not int-like: {cgnc_id}")
            continue

        term = Term.from_triple(
            prefix=PREFIX,
            identifier=cgnc_id,
            name=name,
        )
        term.set_species(identifier="9031", name="Gallus gallus")
        if entrez_id and pd.notna(entrez_id):
            term.append_xref(Reference(prefix="ncbigene", identifier=entrez_id))
        if pd.notna(ensembl_id):
            term.append_xref(Reference(prefix="ensembl", identifier=ensembl_id))
        if synonym_1 and pd.notna(synonym_1):
            term.append_synonym(synonym_1)
        if synoynm_2 and pd.notna(synoynm_2):
            term.append_synonym(synoynm_2)
        yield term


if __name__ == "__main__":
    CGNCGetter.cli()



# -*- coding: utf-8 -*-

"""Tests for getting OBO."""

import unittest
from operator import attrgetter

import obonet

from pyobo import Reference, Synonym, SynonymTypeDef, get_ontology
from pyobo.reader import (
    _extract_definition,
    _extract_synonym,
    iterate_graph_synonym_typedefs,
    iterate_graph_typedefs,
    iterate_node_parents,
    iterate_node_properties,
    iterate_node_relationships,
    iterate_node_synonyms,
    iterate_node_xrefs,
)
from tests.constants import TEST_CHEBI_OBO_PATH, chebi_patch


class TestParseObonet(unittest.TestCase):
    """Test parsing OBO."""

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the test case with a mock ChEBI OBO."""
        cls.ontology = "chebi"
        cls.graph = obonet.read_obo(TEST_CHEBI_OBO_PATH)

    def test_get_graph_typedefs(self):
        """Test getting type definitions from an :mod:`obonet` graph."""
        pairs = {
            (typedef.prefix, typedef.identifier)
            for typedef in iterate_graph_typedefs(self.graph, "chebi")
        }
        self.assertIn(("chebi", "has_part"), pairs)

    def test_get_graph_synonym_typedefs(self):
        """Test getting synonym type definitions from an :mod:`obonet` graph."""
        synonym_typedefs = sorted(
            iterate_graph_synonym_typedefs(self.graph, ontology=self.ontology),
            key=attrgetter("curie"),
        )
        self.assertEqual(
            sorted(
                [
                    SynonymTypeDef(
                        reference=Reference(
                            prefix="chebi", identifier="IUPAC_NAME", name="IUPAC NAME"
                        )
                    ),
                    SynonymTypeDef(
                        reference=Reference(
                            prefix="chebi", identifier="BRAND_NAME", name="BRAND NAME"
                        )
                    ),
                    SynonymTypeDef(
                        reference=Reference(prefix="chebi", identifier="INN", name="INN")
                    ),
                ],
                key=attrgetter("curie"),
            ),
            synonym_typedefs,
        )

    def test_extract_definition(self):
        """Test extracting a definition."""
        expected_text = "Test Text."

        for s, expected_references in [
            (f'"{expected_text}"', []),
            (f'"{expected_text}" []', []),
            (f'"{expected_text}" [PMID:1234]', [Reference(prefix="pubmed", identifier="1234")]),
            (
                f'"{expected_text}" [PMID:1234, PMID:1235]',
                [
                    Reference(prefix="pubmed", identifier="1234"),
                    Reference(prefix="pubmed", identifier="1235"),
                ],
            ),
        ]:
            with self.subTest(s=s):
                actual_text, actual_references = _extract_definition(
                    s, prefix="chebi", identifier="XXX"
                )
                self.assertEqual(expected_text, actual_text)
                self.assertEqual(expected_references, actual_references)

    def test_extract_definition_with_escapes(self):
        """Test extracting a definition with escapes in it."""
        expected_text = """The canonical 3' splice site has the sequence "AG"."""
        s = """"The canonical 3' splice site has the sequence \\"AG\\"." [PMID:1234]"""
        actual_text, actual_references = _extract_definition(
            s, strict=True, prefix="chebi", identifier="XXX"
        )
        self.assertEqual(expected_text, actual_text)
        self.assertEqual([Reference(prefix="pubmed", identifier="1234")], actual_references)

    def test_extract_synonym(self):
        """Test extracting synonym strings."""
        iupac_name = SynonymTypeDef.from_text("IUPAC NAME", lower=False)
        synoynym_typedefs = {
            "IUPAC_NAME": iupac_name,
        }

        for expected_synonym, text in [
            (
                Synonym(
                    name="LTEC I",
                    specificity="EXACT",
                    type=iupac_name,
                    provenance=[Reference(prefix="orphanet", identifier="93938")],
                ),
                '"LTEC I" EXACT IUPAC_NAME [Orphanet:93938]',
            ),
            (
                Synonym(
                    name="LTEC I",
                    specificity="EXACT",
                    provenance=[Reference(prefix="orphanet", identifier="93938")],
                ),
                '"LTEC I" EXACT [Orphanet:93938]',
            ),
            (
                Synonym(
                    name="LTEC I",
                    specificity="EXACT",
                    provenance=[Reference(prefix="orphanet", identifier="93938")],
                ),
                '"LTEC I" [Orphanet:93938]',
            ),
            (
                Synonym(name="LTEC I", specificity="EXACT"),
                '"LTEC I" []',
            ),
        ]:
            with self.subTest(s=text):
                actual_synonym = _extract_synonym(
                    text, synoynym_typedefs, prefix="chebi", identifier="XXX"
                )
                self.assertIsInstance(actual_synonym, Synonym)
                self.assertEqual(expected_synonym, actual_synonym)

    def test_get_node_synonyms(self):
        """Test getting synonyms from a node in a :mod:`obonet` graph."""
        iupac_name = SynonymTypeDef.from_text("IUPAC NAME", lower=False)
        synoynym_typedefs = {
            "IUPAC_NAME": iupac_name,
        }
        data = self.graph.nodes["CHEBI:51990"]
        synonyms = list(
            iterate_node_synonyms(data, synoynym_typedefs, prefix="chebi", identifier="XXX")
        )
        self.assertEqual(1, len(synonyms))
        synonym = synonyms[0]
        self.assertEqual(
            "N,N,N-tributylbutan-1-aminium fluoride", synonym.name, msg="name parsing failed"
        )
        self.assertEqual("EXACT", synonym.specificity, msg="specificity parsing failed")
        self.assertEqual(iupac_name, synonym.type)

    def test_get_node_properties(self):
        """Test getting properties from a node in a :mod:`obonet` graph."""
        data = self.graph.nodes["CHEBI:51990"]
        properties = list(iterate_node_properties(data))
        t_prop = "http://purl.obolibrary.org/obo/chebi/monoisotopicmass"
        self.assertIn(t_prop, {prop for prop, value in properties})
        self.assertEqual(1, sum(prop == t_prop for prop, value in properties))
        value = [value for prop, value in properties if prop == t_prop][0]
        self.assertEqual("261.28318", value)

    def test_get_node_parents(self):
        """Test getting parents from a node in a :mod:`obonet` graph."""
        data = self.graph.nodes["CHEBI:51990"]
        parents = list(iterate_node_parents(data, prefix="chebi", identifier="XXX"))
        self.assertEqual(2, len(parents))
        self.assertEqual({"24060", "51992"}, {parent.identifier for parent in parents})
        self.assertEqual({"chebi"}, {parent.prefix for parent in parents})

    def test_get_node_xrefs(self):
        """Test getting parents from a node in a :mod:`obonet` graph."""
        data = self.graph.nodes["CHEBI:51990"]
        xrefs = list(iterate_node_xrefs(prefix="chebi", data=data))
        self.assertEqual(7, len(xrefs))
        # NOTE the prefixes are remapped by Bioregistry
        self.assertEqual({"pubmed", "cas", "reaxys"}, {xref.prefix for xref in xrefs})
        self.assertEqual(
            {
                ("reaxys", "3570522"),
                ("cas", "429-41-4"),
                ("pubmed", "21142041"),
                ("pubmed", "21517057"),
                ("pubmed", "22229781"),
                ("pubmed", "15074950"),
            },
            {(xref.prefix, xref.identifier) for xref in xrefs},
        )

    def test_get_node_relations(self):
        """Test getting relations from a node in a :mod:`obonet` graph."""
        data = self.graph.nodes["CHEBI:17051"]
        relations = list(iterate_node_relationships(data, prefix="chebi", identifier="XXX"))
        self.assertEqual(1, len(relations))
        typedef, target = relations[0]

        self.assertIsNotNone(target)
        self.assertIsInstance(target, Reference)
        self.assertEqual("chebi", target.prefix)
        self.assertEqual("29228", target.identifier)

        self.assertIsNotNone(typedef)
        self.assertIsInstance(typedef, Reference)
        self.assertEqual("chebi", typedef.prefix)
        self.assertEqual("is_conjugate_base_of", typedef.identifier)


class TestGet(unittest.TestCase):
    """Test generation of OBO objects."""

    def setUp(self) -> None:
        """Set up the test with the mock ChEBI OBO file."""
        with chebi_patch:
            self.ontology = get_ontology("chebi")

    def test_get_terms(self):
        """Test getting an OBO document."""
        terms = list(self.ontology)
        self.assertEqual(18, len(terms))

    def test_get_id_alts_mapping(self):
        """Make sure the alternative ids are mapped properly.

        .. code-block::

            [Term]
            id: CHEBI:16042
            name: halide anion
            alt_id: CHEBI:5605
            alt_id: CHEBI:14384
        """
        id_alts_mapping = self.ontology.get_id_alts_mapping()
        self.assertNotIn("C00462", id_alts_mapping)
        self.assertIn("16042", id_alts_mapping, msg="halide anion alt_id fields not parsed")
        self.assertEqual({"5605", "14384"}, set(id_alts_mapping["16042"]))



# -*- coding: utf-8 -*-

"""KEGG Databases."""

from .genes import KEGGGeneGetter
from .genome import KEGGGenomeGetter
from .pathway import KEGGPathwayGetter

__all__ = [
    "KEGGGeneGetter",
    "KEGGGenomeGetter",
    "KEGGPathwayGetter",
]



# -*- coding: utf-8 -*-

"""DepMap cell lines."""

from typing import Iterable, Optional

import pandas as pd
import pystow

from pyobo import Obo, Reference, Term

__all__ = [
    "get_obo",
    "DepMapGetter",
]

PREFIX = "depmap"
VERSION = "21Q2"


class DepMapGetter(Obo):
    """An ontology representation of the Cancer Dependency Map's cell lines."""

    ontology = bioversions_key = PREFIX
    data_version = VERSION

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise, force=force)


def get_obo(*, force: bool = False) -> Obo:
    """Get DepMap cell lines as OBO."""
    return DepMapGetter(force=force)


def get_url(version: Optional[str] = None) -> str:
    """Get the URL for the given version of the DepMap cell line metadata file.

    :param version: The version of the data
    :returns: The URL as a string for downloading the dat

    .. warning::

        This does not currently take the version into account. Need to write a crawler since data is not easy
        to access.
    """
    #: This is the DepMap Public 21Q2 version. There isn't a way to do this automatically without writing a crawler
    url = "https://ndownloader.figshare.com/files/27902376"
    return url


def _fix_mangled_int(x: str) -> Optional[str]:
    return str(int(float(x))) if pd.notna(x) else None


def iter_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Iterate over DepMap cell line terms."""
    df = ensure(force=force, version=version)
    columns = [
        "DepMap_ID",
        "cell_line_name",
        "stripped_cell_line_name",
        "alias",
        "COSMICID",
        "RRID",
        "WTSI_Master_Cell_ID",
        "Sanger_Model_ID",
    ]
    df["WTSI_Master_Cell_ID"] = df["WTSI_Master_Cell_ID"].map(_fix_mangled_int)
    df["COSMICID"] = df["COSMICID"].map(_fix_mangled_int)
    for identifier, name, sname, aliases, cosmic_id, cellosaurus_id, _wtsi_id, _sanger_id in df[
        columns
    ].values:
        term = Term.from_triple(PREFIX, identifier, name)
        if pd.notna(sname):
            term.append_synonym(sname)
        if pd.notna(aliases):
            for alias in aliases.split(","):
                alias = alias.strip()
                if alias == name:
                    continue
                term.append_synonym(alias)
        if pd.notna(cosmic_id):
            term.append_xref(Reference(prefix="cosmic.cell", identifier=cosmic_id))
        if pd.notna(cellosaurus_id):
            term.append_xref(Reference(prefix="cellosaurus", identifier=cellosaurus_id))

        # WTSI stands for welcome trust sanger institute
        # Not sure where this prefix goes
        # if pd.notna(wtsi_id):
        #    term.append_xref(Reference("sanger", wtsi_id))

        # Not sure what this is
        # if pd.notna(sanger_id):
        #    term.append_xref(Reference("sanger", sanger_id))

        # TODO There's lots of other great ontological information in here. Next time.
        yield term


def ensure(version: str, force: bool = False) -> pd.DataFrame:
    """Ensure and parse the given version of the DepMap cell line metadata."""
    return pystow.ensure_csv(
        "pyobo",
        "raw",
        PREFIX,
        version,
        url=get_url(version=version),
        name="sample_info.tsv",
        force=force,
        read_csv_kwargs=dict(sep=",", dtype=str),
    )


if __name__ == "__main__":
    DepMapGetter.cli()



# -*- coding: utf-8 -*-

"""Converter for Rhea."""

import logging
from typing import Iterable

import pystow

from pyobo.struct import Obo, Reference, Term, TypeDef
from pyobo.utils.path import ensure_df

__all__ = [
    "RheaGetter",
]

logger = logging.getLogger(__name__)
PREFIX = "rhea"

has_lr = TypeDef(Reference(prefix=PREFIX, identifier="has_lr_reaction"))
has_rl = TypeDef(Reference(prefix=PREFIX, identifier="has_rl_reaction"))
has_bi = TypeDef(Reference(prefix=PREFIX, identifier="has_bi_reaction"))


class RheaGetter(Obo):
    """An ontology representation of Rhea's chemical reaction database."""

    ontology = bioversions_key = PREFIX
    typedefs = [has_lr, has_bi, has_rl]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise, force=force)


def get_obo(force: bool = False) -> Obo:
    """Get Rhea as OBO."""
    return RheaGetter(force=force)


def iter_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Iterate over terms in Rhea."""
    terms = {}

    directions = ensure_df(
        PREFIX,
        url="ftp://ftp.expasy.org/databases/rhea/tsv/rhea-directions.tsv",
        version=version,
        force=force,
    )
    for master, lr, rl, bi in directions.values:
        terms[master] = Term(reference=Reference(PREFIX, master))
        terms[lr] = Term(reference=Reference(PREFIX, lr))
        terms[rl] = Term(reference=Reference(PREFIX, rl))
        terms[bi] = Term(reference=Reference(PREFIX, bi))

        terms[master].append_relationship(has_lr, terms[lr])
        terms[master].append_relationship(has_rl, terms[rl])
        terms[master].append_relationship(has_bi, terms[bi])
        terms[lr].append_parent(terms[master])
        terms[rl].append_parent(terms[master])
        terms[bi].append_parent(terms[master])

    hierarchy = ensure_df(
        PREFIX,
        url="ftp://ftp.expasy.org/databases/rhea/tsv/rhea-relationships.tsv",
        version=version,
        force=force,
    )
    for source, relation, target in hierarchy.values:
        if relation != "is_a":
            raise ValueError(f"RHEA unrecognized relation: {relation}")
        terms[source].append_parent(terms[target])

    for xref_prefix, url in [
        ("ecocyc", "rhea2ecocyc"),
        ("kegg.reaction", "rhea2kegg_reaction"),
        ("reactome", "rhea2reactome"),
        ("macie", "rhea2macie"),
        ("metacyc", "rhea2metacyc"),
    ]:
        xref_df = ensure_df(
            PREFIX,
            url=f"ftp://ftp.expasy.org/databases/rhea/tsv/{url}.tsv",
            version=version,
            force=force,
        )
        for rhea_id, _, _, xref_id in xref_df.values:
            if rhea_id not in terms:
                logger.debug(
                    "[%s] could not find %s:%s for xref %s:%s",
                    PREFIX,
                    PREFIX,
                    rhea_id,
                    xref_prefix,
                    xref_id,
                )
                continue
            terms[rhea_id].append_xref(Reference(xref_prefix, xref_id))

    # TODO are EC codes equivalent?
    # TODO uniprot enabled by (RO:0002333)
    # TODO names?

    url = "ftp://ftp.expasy.org/databases/rhea/rdf/rhea.rdf.gz"
    graph = pystow.ensure_rdf(
        "pyobo", "raw", PREFIX, version, url=url, force=force, parse_kwargs=dict(format="xml")
    )
    result = graph.query(
        """
    PREFIX rh:<http://rdf.rhea-db.org/>
    SELECT ?reaction ?reactionId ?reactionLabel WHERE {
      ?reaction rdfs:subClassOf rh:Reaction .
      ?reaction rh:id ?reactionId .
      ?reaction rdfs:label ?reactionLabel .
    }
    """
    )
    for _, identifier, name in result:
        identifier = str(identifier)
        if identifier not in terms:
            logger.debug("isolated element in rdf: rhea:%s ! %s", identifier, name)
            continue
        terms[identifier].reference.name = name

    # TODO participants?

    yield from terms.values()


if __name__ == "__main__":
    RheaGetter.cli()



# -*- coding: utf-8 -*-

"""Importers for selventa terminologies."""

from .schem import SCHEMGetter
from .scomp import SCOMPGetter
from .sdis import SDISGetter
from .sfam import SFAMGetter

__all__ = [
    "SCHEMGetter",
    "SCOMPGetter",
    "SDISGetter",
    "SFAMGetter",
]



# -*- coding: utf-8 -*-

"""Utilities for building paths."""

import logging
from pathlib import Path
from typing import Callable, Optional, Union

import pandas as pd
from pystow.utils import download, name_from_url, read_tarfile_csv

from .misc import cleanup_version
from ..constants import RAW_MODULE

__all__ = [
    "prefix_directory_join",
    "prefix_directory_join",
    "prefix_cache_join",
    "get_prefix_obo_path",
    "ensure_path",
    "ensure_df",
    "ensure_tar_df",
]

logger = logging.getLogger(__name__)

VersionHint = Union[None, str, Callable[[], str]]


def prefix_directory_join(
    prefix: str,
    *parts: str,
    name: Optional[str] = None,
    version: VersionHint = None,
    ensure_exists: bool = True,
) -> Path:
    """Join in the prefix directory."""
    if version is None:
        return RAW_MODULE.join(prefix, *parts, name=name, ensure_exists=ensure_exists)
    if callable(version):
        logger.info("[%s] looking up version", prefix)
        version = version()
        logger.info("[%s] got version %s", prefix, version)
    elif not isinstance(version, str):
        raise TypeError(f"Invalid type: {version} ({type(version)})")
    version = cleanup_version(version, prefix=prefix)
    if version is not None and "/" in version:
        raise ValueError(f"[{prefix}] Can not have slash in version: {version}")
    return RAW_MODULE.join(prefix, version, *parts, name=name, ensure_exists=ensure_exists)


def get_prefix_obo_path(prefix: str, version: VersionHint = None, ext: str = "obo") -> Path:
    """Get the canonical path to the OBO file."""
    return prefix_directory_join(prefix, name=f"{prefix}.{ext}", version=version)


def ensure_path(
    prefix: str,
    *parts: str,
    url: str,
    version: VersionHint = None,
    name: Optional[str] = None,
    force: bool = False,
    error_on_missing: bool = False,
) -> str:
    """Download a file if it doesn't exist."""
    if name is None:
        name = name_from_url(url)

    path = prefix_directory_join(prefix, *parts, name=name, version=version)

    if not path.exists() and error_on_missing:
        raise FileNotFoundError

    download(
        url=url,
        path=path,
        force=force,
    )
    return path.as_posix()


def ensure_df(
    prefix: str,
    *parts: str,
    url: str,
    version: VersionHint = None,
    name: Optional[str] = None,
    force: bool = False,
    sep: str = "\t",
    dtype=str,
    **kwargs,
) -> pd.DataFrame:
    """Download a file and open as a dataframe."""
    _path = ensure_path(prefix, *parts, url=url, version=version, name=name, force=force)
    return pd.read_csv(_path, sep=sep, dtype=dtype, **kwargs)


def ensure_tar_df(
    prefix: str,
    *parts: str,
    url: str,
    inner_path: str,
    version: VersionHint = None,
    path: Optional[str] = None,
    force: bool = False,
    **kwargs,
) -> pd.DataFrame:
    """Download a tar file and open as a dataframe."""
    path = ensure_path(prefix, *parts, url=url, version=version, name=path, force=force)
    return read_tarfile_csv(path, inner_path=inner_path, **kwargs)


def prefix_cache_join(prefix: str, *parts, name: Optional[str], version: VersionHint) -> Path:
    """Ensure the prefix cache is available."""
    return prefix_directory_join(prefix, "cache", *parts, name=name, version=version)



# -*- coding: utf-8 -*-

"""Converter for ComplexPortal."""

import logging
from typing import Iterable, List, Tuple

import pandas as pd
from tqdm.auto import tqdm

from pyobo.resources.ncbitaxon import get_ncbitaxon_name
from pyobo.struct import Obo, Reference, Synonym, Term, from_species, has_part
from pyobo.utils.path import ensure_df

__all__ = [
    "ComplexPortalGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "complexportal"
SPECIES = [
    "10090",
    "10116",
    "1235996",
    "1263720",
    "208964",
    "2697049",
    "284812",
    "3702",
    "559292",
    "562",
    "6239",
    "6523",
    "694009",
    "7227",
    "7787",
    "7788",
    "7955",
    "83333",
    "8355",
    "9031",
    "9606",
    "9615",
    "9823",
    "9913",
    "9940",
    "9986",
]
DTYPE = {
    "taxonomy_id": str,
}


def _parse_members(s) -> List[Tuple[Reference, str]]:
    if pd.isna(s):
        return []

    rv = []
    for member in s.split("|"):
        entity_id, count = member.split("(")
        count = count.rstrip(")")
        if ":" in entity_id:
            prefix, identifier = entity_id.split(":", 1)
        else:
            prefix, identifier = "uniprot", entity_id
        rv.append((Reference(prefix=prefix, identifier=identifier), count))
    return rv


def _parse_xrefs(s) -> List[Tuple[Reference, str]]:
    if pd.isna(s):
        return []

    rv = []
    for xref in s.split("|"):
        xref = xref.replace("protein ontology:PR:", "PR:")
        xref = xref.replace("protein ontology:PR_", "PR:")
        try:
            xref_curie, note = xref.split("(")
        except ValueError:
            logger.warning("xref missing (: %s", xref)
            continue
        note = note.rstrip(")")
        try:
            reference = Reference.from_curie(xref_curie)
        except ValueError:
            logger.warning("can not parse CURIE: %s", xref)
            continue
        if reference is None:
            logger.warning("reference is None after parsing: %s", xref)
            continue
        rv.append((reference, note))
    return rv


class ComplexPortalGetter(Obo):
    """An ontology representation of the Complex Portal."""

    bioversions_key = ontology = PREFIX
    typedefs = [from_species, has_part]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return get_terms(version=self._version_or_raise)


def get_obo(force: bool = False) -> Obo:
    """Get the ComplexPortal OBO."""
    return ComplexPortalGetter(force=force)


def get_df(version: str, force: bool = False) -> pd.DataFrame:
    """Get a combine ComplexPortal dataframe."""
    url_base = f"ftp://ftp.ebi.ac.uk/pub/databases/intact/complex/{version}/complextab"
    dfs = [
        ensure_df(
            PREFIX,
            url=f"{url_base}/{ncbitaxonomy_id}.tsv",
            version=version,
            na_values={"-"},
            header=0,
            dtype=str,
            force=force,
        )
        for ncbitaxonomy_id in SPECIES
    ]
    return pd.concat(dfs)


def get_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Get ComplexPortal terms."""
    df = get_df(version=version, force=force)
    df.rename(
        inplace=True,
        columns={
            "Aliases for complex": "aliases",
            "Identifiers (and stoichiometry) of molecules in complex": "members",
            "Taxonomy identifier": "taxonomy_id",
            "Cross references": "xrefs",
            "Description": "definition",
            "Recommended name": "name",
            "#Complex ac": "complexportal_id",
        },
    )

    df["aliases"] = df["aliases"].map(lambda s: s.split("|") if pd.notna(s) else [])
    df["members"] = df["members"].map(_parse_members)
    df["xrefs"] = df["xrefs"].map(_parse_xrefs)

    df["taxonomy_name"] = df["taxonomy_id"].map(get_ncbitaxon_name)

    slim_df = df[
        [
            "complexportal_id",
            "name",
            "definition",
            "aliases",
            "xrefs",
            "taxonomy_id",
            "taxonomy_name",
            "members",
        ]
    ]
    it = tqdm(slim_df.values, total=len(slim_df.index), desc=f"mapping {PREFIX}")
    unhandled_xref_type = set()
    for (
        complexportal_id,
        name,
        definition,
        aliases,
        xrefs,
        taxonomy_id,
        taxonomy_name,
        members,
    ) in it:
        synonyms = [Synonym(name=alias) for alias in aliases]
        _xrefs = []
        provenance = []
        for reference, note in xrefs:
            if note == "identity":
                _xrefs.append(reference)
            elif note == "see-also" and reference.prefix == "pubmed":
                provenance.append(reference)
            elif (note, reference.prefix) not in unhandled_xref_type:
                logger.debug(f"unhandled xref type: {note} / {reference.prefix}")
                unhandled_xref_type.add((note, reference.prefix))

        term = Term(
            reference=Reference(prefix=PREFIX, identifier=complexportal_id, name=name),
            definition=definition.strip() if pd.notna(definition) else None,
            synonyms=synonyms,
            xrefs=_xrefs,
            provenance=provenance,
        )
        term.set_species(identifier=taxonomy_id, name=taxonomy_name)

        for reference, _count in members:
            term.append_relationship(has_part, reference)

        yield term


if __name__ == "__main__":
    ComplexPortalGetter.cli()



# -*- coding: utf-8 -*-

"""Convert PFAM Clans to OBO."""

from typing import Iterable

from tqdm.auto import tqdm

from .pfam import get_pfam_clan_df
from ..struct import Obo, Reference, Term

__all__ = [
    "PfamClanGetter",
]

PREFIX = "pfam.clan"


class PfamClanGetter(Obo):
    """An ontology representation of Pfam's protein clan nomenclature."""

    ontology = PREFIX
    bioversions_key = "pfam"

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise, force=force)


def get_obo(force: bool = False) -> Obo:
    """Get PFAM Clans as OBO."""
    return PfamClanGetter(force=force)


# TODO could get definitions from ftp://ftp.ebi.ac.uk/pub/databases/Pfam/releases/Pfam33.0/Pfam-C.gz


def iter_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Iterate PFAM clan terms."""
    df = get_pfam_clan_df(version=version, force=force)
    df = df[df["clan_id"].notna()]
    df = df[["clan_id", "clan_name"]].drop_duplicates()
    it = tqdm(df.values, total=len(df.index), desc=f"mapping {PREFIX}")
    for identifier, name in it:
        yield Term(
            reference=Reference(prefix=PREFIX, identifier=identifier, name=name),
        )


if __name__ == "__main__":
    get_obo().write_default()



# -*- coding: utf-8 -*-

"""Pipeline for extracting all xrefs from OBO documents available."""

import gzip
import itertools as itt
import logging
from typing import Iterable, Optional, Tuple, cast

import bioregistry
import networkx as nx
import pandas as pd
from tqdm.auto import tqdm

from .sources import iter_xref_plugins
from .. import get_xrefs_df
from ..api import (
    get_id_definition_mapping,
    get_id_name_mapping,
    get_id_species_mapping,
    get_id_synonyms_mapping,
    get_id_to_alts,
    get_metadata,
    get_properties_df,
    get_relations_df,
    get_typedef_df,
)
from ..constants import SOURCE_ID, SOURCE_PREFIX, TARGET_ID, TARGET_PREFIX
from ..getters import iter_helper, iter_helper_helper
from ..sources import ncbigene, pubchem
from ..utils.path import ensure_path

logger = logging.getLogger(__name__)


# TODO a normal graph can easily be turned into a directed graph where each
#  edge points from low priority to higher priority, then the graph can
#  be reduced to a set of star graphs and ultimately to a single dictionary


def get_graph_from_xref_df(df: pd.DataFrame) -> nx.Graph:
    """Generate a graph from the mappings dataframe."""
    rv = nx.Graph()

    it = itt.chain(
        df[[SOURCE_PREFIX, SOURCE_ID]].drop_duplicates().values,
        df[[TARGET_PREFIX, TARGET_ID]].drop_duplicates().values,
    )
    it = tqdm(it, desc="loading curies", unit_scale=True)
    for prefix, identifier in it:
        rv.add_node(_to_curie(prefix, identifier), prefix=prefix, identifier=identifier)

    it = tqdm(df.values, total=len(df.index), desc="loading xrefs", unit_scale=True)
    for source_ns, source_id, target_ns, target_id, provenance in it:
        rv.add_edge(
            _to_curie(source_ns, source_id),
            _to_curie(target_ns, target_id),
            provenance=provenance,
        )

    return rv


def _to_curie(prefix: str, identifier: str) -> str:
    return f"{prefix}:{identifier}"


def _iter_ncbigene(left, right):
    ncbi_path = ensure_path(ncbigene.PREFIX, url=ncbigene.GENE_INFO_URL)
    with gzip.open(ncbi_path, "rt") as file:
        next(file)  # throw away the header
        for line in tqdm(
            file, desc=f"extracting {ncbigene.PREFIX}", unit_scale=True, total=27_000_000
        ):
            line = line.strip().split("\t")
            yield ncbigene.PREFIX, line[left], line[right]


def _iter_metadata(**kwargs):
    for prefix, data in iter_helper_helper(get_metadata, **kwargs):
        version = data["version"]
        tqdm.write(f"[{prefix}] using version {version}")
        yield prefix, version, data["date"], bioregistry.is_deprecated(prefix)


def _iter_names(leave: bool = False, **kwargs) -> Iterable[Tuple[str, str, str]]:
    """Iterate over all prefix-identifier-name triples we can get.

    :param leave: should the tqdm be left behind?
    """
    yield from iter_helper(get_id_name_mapping, leave=leave, **kwargs)
    yield from _iter_ncbigene(1, 2)

    pcc_path = pubchem._ensure_cid_name_path()
    with gzip.open(pcc_path, mode="rt", encoding="ISO-8859-1") as file:
        for line in tqdm(
            file, desc=f"extracting {pubchem.PREFIX}", unit_scale=True, total=103_000_000
        ):
            identifier, name = line.strip().split("\t", 1)
            yield pubchem.PREFIX, identifier, name


def _iter_species(leave: bool = False, **kwargs) -> Iterable[Tuple[str, str, str]]:
    """Iterate over all prefix-identifier-species triples we can get."""
    yield from iter_helper(get_id_species_mapping, leave=leave, **kwargs)
    # TODO ncbigene


def _iter_definitions(leave: bool = False, **kwargs) -> Iterable[Tuple[str, str, str]]:
    """Iterate over all prefix-identifier-descriptions triples we can get."""
    yield from iter_helper(get_id_definition_mapping, leave=leave, **kwargs)
    yield from _iter_ncbigene(1, 8)


def _iter_alts(
    leave: bool = False, strict: bool = True, **kwargs
) -> Iterable[Tuple[str, str, str]]:
    for prefix, identifier, alts in iter_helper(
        get_id_to_alts, leave=leave, strict=strict, **kwargs
    ):
        for alt in alts:
            yield prefix, identifier, alt


def _iter_synonyms(leave: bool = False, **kwargs) -> Iterable[Tuple[str, str, str]]:
    """Iterate over all prefix-identifier-synonym triples we can get.

    :param leave: should the tqdm be left behind?
    """
    for prefix, identifier, synonyms in iter_helper(get_id_synonyms_mapping, leave=leave, **kwargs):
        for synonym in synonyms:
            yield prefix, identifier, synonym


def _iter_typedefs(**kwargs) -> Iterable[Tuple[str, str, str, str]]:
    """Iterate over all prefix-identifier-name triples we can get."""
    for prefix, df in iter_helper_helper(get_typedef_df, **kwargs):
        for t in df.values:
            if all(t):
                yield cast(Tuple[str, str, str, str], (prefix, *t))


def _iter_relations(**kwargs) -> Iterable[Tuple[str, str, str, str, str, str]]:
    for prefix, df in iter_helper_helper(get_relations_df, **kwargs):
        for t in df.values:
            if all(t):
                yield cast(Tuple[str, str, str, str, str, str], (prefix, *t))


def _iter_properties(**kwargs) -> Iterable[Tuple[str, str, str, str]]:
    for prefix, df in iter_helper_helper(get_properties_df, **kwargs):
        for t in df.values:
            if all(t):
                yield cast(Tuple[str, str, str, str], (prefix, *t))


def _iter_xrefs(
    *,
    force: bool = False,
    use_tqdm: bool = True,
    skip_below: Optional[str] = None,
    strict: bool = True,
    **kwargs,
) -> Iterable[Tuple[str, str, str, str, str]]:
    it = iter_helper_helper(
        get_xrefs_df,
        use_tqdm=use_tqdm,
        force=force,
        skip_below=skip_below,
        strict=strict,
        **kwargs,
    )
    for prefix, df in it:
        df.dropna(inplace=True)
        for row in df.values:
            if any(not element for element in row):
                continue
            yield cast(Tuple[str, str, str, str, str], (prefix, *row, prefix))
    for df in iter_xref_plugins(skip_below=skip_below):
        df.dropna(inplace=True)
        yield from tqdm(df.values, leave=False, total=len(df.index), unit_scale=True)



# -*- coding: utf-8 -*-

"""Data structures for OBO."""

from .reference import Reference  # noqa: F401
from .struct import (  # noqa: F401
    Obo,
    Synonym,
    SynonymSpecificities,
    SynonymSpecificity,
    SynonymTypeDef,
    Term,
    make_ad_hoc_ontology,
)
from .typedef import (  # noqa: F401
    RelationHint,
    TypeDef,
    enables,
    from_species,
    gene_product_member_of,
    get_reference_tuple,
    has_gene_product,
    has_member,
    has_part,
    has_participant,
    is_a,
    member_of,
    orthologous,
    part_of,
    participates_in,
    species_specific,
    superclass_of,
    transcribes_to,
    translates_to,
)



# -*- coding: utf-8 -*-

"""Converter for InterPro."""

from collections import defaultdict
from typing import DefaultDict, Iterable, List, Mapping, Set, Tuple

from .utils import get_go_mapping
from ..struct import Obo, Reference, Term
from ..struct.typedef import enables, has_member
from ..utils.io import multisetdict
from ..utils.path import ensure_df, ensure_path

__all__ = [
    "InterProGetter",
]

PREFIX = "interpro"

#: Data source for protein-interpro mappings
INTERPRO_PROTEIN_COLUMNS = [
    "uniprot_id",
    "interpro_id",
    "interpro_name",
    "xref",  # either superfamily, gene family gene scan, PFAM, TIGERFAM
    "start",  # int
    "end",  # int
]


class InterProGetter(Obo):
    """An ontology representation of InterPro."""

    ontology = bioversions_key = PREFIX
    typedefs = [enables, has_member]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over InterPro terms."""
        return iter_terms(version=self._version_or_raise, force=force)


def get_obo(force: bool = False) -> Obo:
    """Get InterPro as OBO."""
    return InterProGetter(force=force)


def iter_terms(*, version: str, proteins: bool = False, force: bool = False) -> Iterable[Term]:
    """Get InterPro terms."""
    parents = get_interpro_tree(version=version, force=force)
    interpro_to_gos = get_interpro_go_df(version=version, force=force)
    interpro_to_proteins = (
        get_interpro_to_proteins_df(version=version, force=force) if proteins else {}
    )

    entries_df = ensure_df(
        PREFIX,
        url=f"https://ftp.ebi.ac.uk/pub/databases/interpro/releases/{version}/entry.list",
        name="entries.tsv",
        skiprows=1,
        names=("ENTRY_AC", "ENTRY_TYPE", "ENTRY_NAME"),
        version=version,
        force=force,
    )

    references = {
        identifier: Reference(prefix=PREFIX, identifier=identifier, name=name)
        for identifier, _, name in entries_df.values
    }

    for identifier, entry_type, _ in entries_df.values:
        term = Term(
            reference=references[identifier],
            parents=[references[parent_id] for parent_id in parents.get(identifier, [])],
        )
        for go_id, go_name in interpro_to_gos.get(identifier, []):
            term.append_relationship(
                enables, Reference(prefix="go", identifier=go_id, name=go_name)
            )
        term.append_property("type", entry_type)
        for uniprot_id in interpro_to_proteins.get(identifier, []):
            term.append_relationship(has_member, Reference(prefix="uniprot", identifier=uniprot_id))
        yield term


def get_interpro_go_df(version: str, force: bool = False) -> Mapping[str, Set[Tuple[str, str]]]:
    """Get InterPro to Gene Ontology molecular function mapping."""
    url = f"https://ftp.ebi.ac.uk/pub/databases/interpro/releases/{version}/interpro2go"
    path = ensure_path(PREFIX, url=url, name="interpro2go.tsv", version=version, force=force)
    return get_go_mapping(path, prefix=PREFIX)


def get_interpro_tree(version: str, force: bool = False):
    """Get InterPro Data source."""
    url = f"https://ftp.ebi.ac.uk/pub/databases/interpro/releases/{version}/ParentChildTreeFile.txt"
    path = ensure_path(PREFIX, url=url, version=version, force=force)
    with open(path) as f:
        return _parse_tree_helper(f)


def _parse_tree_helper(lines: Iterable[str]):
    rv1: DefaultDict[str, List[str]] = defaultdict(list)
    previous_depth, previous_id = 0, ""
    stack = [previous_id]

    for line in lines:
        depth = _count_front(line)
        parent_id, _ = line[depth:].split("::", maxsplit=1)

        if depth == 0:
            stack.clear()
            stack.append(parent_id)
        else:
            if depth > previous_depth:
                stack.append(previous_id)

            elif depth < previous_depth:
                del stack[-1]

            child_id = stack[-1]
            rv1[child_id].append(parent_id)  # type:ignore

        previous_depth, previous_id = depth, parent_id

    rv2 = defaultdict(list)
    for k, vs in rv1.items():
        for v in vs:
            rv2[v].append(k)
    return dict(rv2)


def _count_front(s: str) -> int:
    """Count the number of leading dashes on a string."""
    for position, element in enumerate(s):
        if element != "-":
            return position
    raise ValueError


def get_interpro_to_proteins_df(version: str, force: bool = False):
    """Get InterPro to Protein dataframe."""
    url = f"https://ftp.ebi.ac.uk/pub/databases/interpro/releases/{version}/protein2ipr.dat.gz"
    df = ensure_df(
        PREFIX,
        url=url,
        compression="gzip",
        usecols=[0, 1, 3],
        names=INTERPRO_PROTEIN_COLUMNS,
        version=version,
        force=force,
    )
    return multisetdict((interpro_id, uniprot_id) for uniprot_id, interpro_id in df.values)


if __name__ == "__main__":
    InterProGetter.cli()



# -*- coding: utf-8 -*-

"""Tests for famplex."""

import unittest

from pyobo.xrefdb.sources.famplex import _get_famplex_df


class TestFamplex(unittest.TestCase):
    """Tests for famplex."""

    def test_mapping(self):
        """Test geting the mapping df."""
        df = _get_famplex_df(force=True)
        self.assertIsNotNone(df)



# -*- coding: utf-8 -*-

"""PyOBO's Gilda utilities."""

import itertools as itt
import logging
from typing import Iterable, List, Optional, Tuple, Type, Union

import bioregistry
import gilda.api
import gilda.term
from gilda.grounder import Grounder
from gilda.process import normalize
from tqdm.auto import tqdm

from pyobo import (
    get_id_name_mapping,
    get_id_species_mapping,
    get_id_synonyms_mapping,
    get_ids,
)
from pyobo.getters import NoBuild
from pyobo.utils.io import multidict

__all__ = [
    "iter_gilda_prediction_tuples",
    "get_grounder",
    "get_gilda_terms",
]

logger = logging.getLogger(__name__)


_STATUSES = {"curated": 1, "name": 2, "synonym": 3, "former_name": 4}


def filter_out_duplicates(terms: List[gilda.term.Term]) -> List[gilda.term.Term]:
    """Filter out duplicates."""
    # TODO import from gilda.term import filter_out_duplicates when it gets moved,
    #  see https://github.com/indralab/gilda/pull/103
    logger.debug("filtering %d terms for uniqueness", len(terms))
    new_terms: List[gilda.term.Term] = [
        min(terms_group, key=_status_key)
        for _, terms_group in itt.groupby(sorted(terms, key=_term_key), key=_term_key)
    ]
    # Re-sort the terms
    new_terms = sorted(new_terms, key=lambda x: (x.text, x.db, x.id))
    logger.debug("got %d unique terms.", len(new_terms))
    return new_terms


def _status_key(term: gilda.term.Term) -> int:
    return _STATUSES[term.status]


def _term_key(term: gilda.term.Term) -> Tuple[str, str, str]:
    return term.db, term.id, term.text


def iter_gilda_prediction_tuples(
    prefix: str,
    relation: str = "skos:exactMatch",
    *,
    grounder: Optional[Grounder] = None,
    identifiers_are_names: bool = False,
    strict: bool = False,
) -> Iterable[Tuple[str, str, str, str, str, str, str, str, float]]:
    """Iterate over prediction tuples for a given prefix."""
    if grounder is None:
        grounder = gilda.api.grounder
    id_name_mapping = get_id_name_mapping(prefix, strict=strict)
    it = tqdm(
        id_name_mapping.items(), desc=f"[{prefix}] gilda tuples", unit_scale=True, unit="name"
    )
    for identifier, name in it:
        for scored_match in grounder.ground(name):
            target_prefix = scored_match.term.db.lower()
            yield (
                prefix,
                normalize_identifier(prefix, identifier),
                name,
                relation,
                target_prefix,
                normalize_identifier(target_prefix, scored_match.term.id),
                scored_match.term.entry_name,
                "semapv:LexicalMatching",
                round(scored_match.score, 3),
            )

    if identifiers_are_names:
        it = tqdm(get_ids(prefix), desc=f"[{prefix}] gilda tuples", unit_scale=True, unit="id")
        for identifier in it:
            for scored_match in grounder.ground(identifier):
                target_prefix = scored_match.term.db.lower()
                yield (
                    prefix,
                    normalize_identifier(prefix, identifier),
                    identifier,
                    relation,
                    target_prefix,
                    normalize_identifier(target_prefix, scored_match.term.id),
                    scored_match.term.entry_name,
                    "semapv:LexicalMatching",
                    scored_match.score,
                )


def normalize_identifier(prefix: str, identifier: str) -> str:
    """Normalize the identifier."""
    resource = bioregistry.get_resource(prefix)
    if resource is None:
        raise KeyError
    return resource.miriam_standardize_identifier(identifier) or identifier


def get_grounder(
    prefixes: Union[str, Iterable[str]],
    unnamed: Optional[Iterable[str]] = None,
    grounder_cls: Optional[Type[Grounder]] = None,
    versions: Union[None, str, Iterable[Union[str, None]]] = None,
    strict: bool = True,
) -> Grounder:
    """Get a Gilda grounder for the given prefix(es)."""
    unnamed = set() if unnamed is None else set(unnamed)
    if isinstance(prefixes, str):
        prefixes = [prefixes]
    else:
        prefixes = list(prefixes)
    if versions is None:
        versions = [None] * len(prefixes)
    elif isinstance(versions, str):
        versions = [versions]
    else:
        versions = list(versions)
    if len(prefixes) != len(versions):
        raise ValueError

    terms: List[gilda.term.Term] = []
    for prefix, version in zip(prefixes, versions):
        try:
            p_terms = list(
                get_gilda_terms(
                    prefix, identifiers_are_names=prefix in unnamed, version=version, strict=strict
                )
            )
        except NoBuild:
            continue
        else:
            terms.extend(p_terms)
    terms = filter_out_duplicates(terms)
    terms_dict = multidict((term.norm_text, term) for term in terms)
    if grounder_cls is None:
        return Grounder(terms_dict)
    else:
        return grounder_cls(terms_dict)


def get_gilda_terms(
    prefix: str,
    identifiers_are_names: bool = False,
    version: Optional[str] = None,
    strict: bool = True,
) -> Iterable[gilda.term.Term]:
    """Get gilda terms for the given namespace."""
    id_to_name = get_id_name_mapping(prefix, version=version, strict=strict)
    id_to_species = get_id_species_mapping(prefix, version=version, strict=strict)

    it = tqdm(id_to_name.items(), desc=f"[{prefix}] mapping", unit_scale=True, unit="name")
    for identifier, name in it:
        yield gilda.term.Term(
            norm_text=normalize(name),
            text=name,
            db=prefix,
            id=identifier,
            entry_name=name,
            status="name",
            source=prefix,
            organism=id_to_species.get(identifier),
        )

    id_to_synonyms = get_id_synonyms_mapping(prefix, version=version)
    if id_to_synonyms:
        it = tqdm(
            id_to_synonyms.items(), desc=f"[{prefix}] mapping", unit_scale=True, unit="synonym"
        )
        for identifier, synonyms in it:
            name = id_to_name[identifier]
            for synonym in synonyms:
                yield gilda.term.Term(
                    norm_text=normalize(synonym),
                    text=synonym,
                    db=prefix,
                    id=identifier,
                    entry_name=name,
                    status="synonym",
                    source=prefix,
                    organism=id_to_species.get(identifier),
                )

    if identifiers_are_names:
        it = tqdm(get_ids(prefix), desc=f"[{prefix}] mapping", unit_scale=True, unit="id")
        for identifier in it:
            yield gilda.term.Term(
                norm_text=normalize(identifier),
                text=identifier,
                db=prefix,
                id=identifier,
                entry_name=None,
                status="identifier",
                source=prefix,
                organism=id_to_species.get(identifier),
            )
